# Copyright (C) 2002-2022 S[&]T, The Netherlands.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import wx
import wx.grid

import coda
import re
import os
import sys
import math
import struct
import weakref

from . import model
from . import renderers


# CONSTANTS
_maxLineLength = 256


# EXCEPTION HIERARCHY
class ProductBrowserError(Exception):
    """Generalization of all ProductBrowser errors"""
    pass


class PathError(ProductBrowserError):

    def __init__(self, str):
        ProductBrowserError.__init__(self)
        self.str = str

    def __str__(self):
        return "PathError: " + self.str


# CUSTOM EVENTS
wxEVT_ITEM_SELECTED = wx.NewEventType()
EVT_ITEM_SELECTED = wx.PyEventBinder(wxEVT_ITEM_SELECTED)
wxEVT_ITEM_DESELECTED = wx.NewEventType()
EVT_ITEM_DESELECTED = wx.PyEventBinder(wxEVT_ITEM_DESELECTED)
wxEVT_CHILD_FOCUS = wx.NewEventType()
EVT_CHILD_FOCUS = wx.PyEventBinder(wxEVT_CHILD_FOCUS)


class ItemSelectedEvent(wx.PyCommandEvent):

    eventType = wxEVT_ITEM_SELECTED

    def __init__(self, id, index):
        wx.PyCommandEvent.__init__(self, ItemSelectedEvent.eventType, id)
        self.index = index

    def Clone(self):
        tmp = self.__class__(self.GetId(), self.index)
        tmp.SetEventObject(self.GetEvenObject())
        return tmp


class ItemDeselectedEvent(wx.PyCommandEvent):

    eventType = wxEVT_ITEM_DESELECTED

    def __init__(self, id, index):
        wx.PyCommandEvent.__init__(self, ItemDeselectedEvent.eventType, id)
        self.index = index

    def Clone(self):
        tmp = self.__class__(self.GetId(), self.index)
        tmp.SetEventObject(self.GetEvenObject())
        return tmp


class ChildFocusEvent(wx.PyCommandEvent):
    eventType = wxEVT_CHILD_FOCUS

    def __init__(self, id):
        wx.PyCommandEvent.__init__(self, ChildFocusEvent.eventType, id)

    def Clone(self):
        tmp = self.__class__(self.GetId())
        tmp.SetEventObject(self.GetEvenObject())
        return tmp


class Observable(object):

    def __init__(self):
        self.observers = []

    def register(self, observer):
        assert hasattr(observer, "notify"), "Observer should have a callable attribute \"notify\""
        assert callable(observer.notify), "Observer should have a callable attribute \"notify\""
        self.observers.append(observer)

    def unregister(self, observer):
        self.observers.remove(observer)

    def notifyObservers(self):
        for observer in self.observers:
            observer.notify(self)


class ButtonGroup(object):

    def __init__(self):
        self.map = {}
        self.buttons = []
        self.selected = -1

    def AppendButton(self, button):
        self.map[button.GetId()] = len(self.buttons)
        self.buttons.append(button)

        if self.selected == -1:
            self.selected = 0
            button.SetValue(True)

    def GetSelected(self):
        return self.selected

    def SetSelected(self, index):
        assert (index >= 0) and (index < len(self.buttons)), "Button index out of range"
        if self.selected != -1:
            self.buttons[self.selected].SetValue(False)
            self.selected = index
            self.buttons[self.selected].SetValue(True)

    def ProcessToggleEvent(self, button):
        try:
            index = self.map[button.GetId()]
        except KeyError:
            return False

        if index == self.selected:
            button.SetValue(True)
            return False
        else:
            self.buttons[self.selected].SetValue(False)
            self.selected = index
            return True


class NavigationController(Observable):

    def __init__(self, frame):
        Observable.__init__(self)

        self.frame = frame
        self.view = None
        self.rootNode = None
        self.productRootNode = None
        self.selectedNode = None

        self.autoExpanding = False
        self.autoPath = None
        self.autoIndex = -1

        self.ignoreFocus = False

    def Initialize(self, view, cursor):
        self.view = view

        try:
            self.productRootNode = model.NodeFactory(cursor)
            self.productRootNode.cursor = cursor
            self.productRootNode.name = "|product|"
            self.productRootNode.initialize()

        except coda.CodacError as ex:
            self.productRootNode = None
            message = wx.MessageDialog(self.frame, "Error reading product root node.\nPlease check product sanity, "
                                       "e.g. using codacheck.\n\nCODA error message: \"[CODA] %s\"" % str(ex),
                                       "CODA error", style=wx.OK | wx.ICON_ERROR | wx.STAY_ON_TOP)
            message.ShowModal()
            self.view.LayoutAndRefresh()
        else:
            self.rootNode = model.RecordNode()
            self.rootNode.name = "|root|"
            self.rootNode._count = 1
            self.rootNode._fields = [self.productRootNode]

            # set parent of productRootNode
            self.productRootNode.parent = weakref.proxy(self.rootNode)

            rootView = self.view.ViewFactory(self.rootNode)
            self.view.PushView(rootView)
            self.view.LayoutAndRefresh()

            rootView.SelectItem(0)

    def AutoExpand(self, node):
        assert (self.autoIndex >= 0) and (self.autoIndex <= len(self.autoPath) - 1), "AutoIndex out of range"

        if self.autoIndex == len(self.autoPath) - 1:
            # reached the end of the path.
            return False
        elif node.name != self.autoPath[self.autoIndex]:
            # mismatch, can occur when a user clicks a node during auto-expansion
            return False
        else:
            # try to find next path element.
            view = self.view.GetView(self.autoIndex + 1)
            index = view.Find(self.autoPath[self.autoIndex + 1])

            if index < 0:
                return False
            else:
                # found it.
                self.autoIndex += 1
                wx.CallAfter(view.SelectItem, index)
                return True

    def OnItemDeselected(self, view, parent, index):
        # pop all views to the right of this view
        self.view.PopUntil(view)

        # unfortunately, this notifyObservers() call is necessary because
        # it is impossible to tell if a user is deselecting a node or changing
        # a selection. in the first case, no subsequent select event will be send
        # which would leave the interface in an inconsistent state without the calls
        # below.
        self.selectedNode = parent

        # ensure a path length of either 0 or at least 2 on win32.
        if self.selectedNode == self.rootNode:
            self.selectedNode = None

        self.ignoreFocus = True
        self.notifyObservers()

        # give focus to parent view.
        index = self.view.GetViewIndex(view)
        if index >= 1:
            parentView = self.view.GetView(index - 1)
            parentView.SetFocus()

        self.ignoreFocus = False

    def OnItemSelected(self, view, parent, index):
        self.ignoreFocus = True

        # try to access the selected node.
        try:
            node = parent[index]
        except model.CorruptProductError as ex:
            message = wx.MessageDialog(self.frame, "Error reading array element.\nPlease check product sanity, "
                                       "e.g. using codacheck.\n\nCODA error message: \"%s\"" % str(ex),
                                       "CODA error", style=wx.OK | wx.ICON_ERROR | wx.STAY_ON_TOP)
            message.ShowModal()

            if sys.platform != "win32":
                self.view.PopUntil(view)

            self.autoExpanding = False
            self.selectedNode = parent
            self.notifyObservers()
            if sys.platform in ["darwin", "win32"]:
                view.SetFocus()
            self.view.ScrollToEnd()
            self.ignoreFocus = False
            return

        # ignore the selection if the node is already selected.
        if node == self.selectedNode and not self.autoExpanding:
            self.ignoreFocus = False
            return

        # pop all the views to the right of the current view. on win32 this has already
        # happened due to the deselect event.
        if sys.platform != "win32":
            self.view.PopUntil(view)

        # if the selected node is a record, try to retrieve information
        # about its fields.
        if node.type == model.TYPE_RECORD and node.dirty():
            try:
                model.RetrieveFieldInfo(node)
            except model.CorruptProductError as ex:
                message = wx.MessageDialog(self.frame, "Error reading record field information.\nPlease check product "
                                           "sanity, e.g. using codacheck.\n\nCODA error message: \"%s\"" % str(ex),
                                           "CODA error", style=wx.OK | wx.ICON_ERROR | wx.STAY_ON_TOP)
                message.ShowModal()

                self.autoExpanding = False
                self.selectedNode = node
                self.notifyObservers()
                if sys.platform in ["darwin", "win32"]:
                    view.SetFocus()
                self.view.ScrollToEnd()
                self.ignoreFocus = False
                return

        # create a new view for the selected node.
        self.view.PushView(self.view.ViewFactory(node))
        self.view.LayoutAndRefresh()

        # if needed, try to auto expanded the next node on the path.
        # NB. AutoExpand() may change the value of autoExpanding.
        if self.autoExpanding:
            self.autoExpanding = self.AutoExpand(node)

        if not self.autoExpanding:
            # the end of the path was reached, the next path element could
            # not be found, or the user manually selected a node.
            self.selectedNode = node
            self.notifyObservers()
            if sys.platform in ["darwin", "win32"]:
                view.SetFocus()
            self.view.ScrollToEnd()
            self.ignoreFocus = False

    def OnChildFocus(self, view, parent):
        if not self.ignoreFocus:
            self.ignoreFocus = True

            index = view.GetSelectedIndex()
            if index >= 0:
                try:
                    selectedNode = parent[index]
                except model.CorruptProductError:
                    selectedNode = parent
            else:
                selectedNode = None

            if selectedNode and self.selectedNode and selectedNode != self.selectedNode:
                self.selectedNode = selectedNode
                self.notifyObservers()
                if sys.platform in ["darwin", "win32"]:
                    view.SetFocus()

            self.ignoreFocus = False

    def Back(self):
        if self.autoExpanding:
            return

        if not self.selectedNode:
            return

        self.ignoreFocus = True

        index = len(self.selectedNode.path()) - 2
        view = self.view.GetView(index)

        if index > 0:
            previousView = self.view.GetView(index - 1)
            parent = previousView.GetNode()
            selectedIndex = previousView.GetSelectedIndex()
            if selectedIndex >= 0:
                try:
                    node = parent[selectedIndex]
                except model.CorruptProductError:
                    view.SetFocus()
                    self.ignoreFocus = False
                    return

                if node and self.selectedNode and node != self.selectedNode:
                    self.selectedNode = node
                    self.notifyObservers()

            previousView.SetFocus()
            self.ignoreFocus = False
        else:
            view.SetFocus()
            self.ignoreFocus = False

    def Forward(self):
        if self.autoExpanding:
            return

        if not self.selectedNode:
            return

        self.ignoreFocus = True

        index = len(self.selectedNode.path()) - 2
        view = self.view.GetView(index)

        if index < self.view.GetViewCount() - 1:
            nextView = self.view.GetView(index + 1)
            parent = nextView.GetNode()
            selectedIndex = nextView.GetSelectedIndex()
            if selectedIndex >= 0:
                try:
                    node = parent[selectedIndex]
                except model.CorruptProductError:
                    view.SetFocus()
                    self.ignoreFocus = False
                    return

                if node and self.selectedNode and node != self.selectedNode:
                    self.selectedNode = node
                    self.notifyObservers()

            nextView.SetFocus()
            self.ignoreFocus = False
        else:
            view.SetFocus()
            self.ignoreFocus = False

    def ParsePath(self, pathString):
        if pathString.endswith('['):
            raise PathError("path cannot end with \"[\"")

        fieldRegEx = re.compile(r"([^\\[\\]/]+)")
        dimensionRegEx = re.compile(r"([\s0-9\s]+)")

        path = []
        index = 0
        if pathString[0] == '/':
            # consume starting /
            index += 1
        while index < len(pathString):
            fieldDescriptor = fieldRegEx.match(pathString, index)
            if not (fieldDescriptor is None):
                (fieldName,) = fieldDescriptor.groups()

                if fieldName is None:
                    raise PathError("invalid field name")

                path.append(fieldName)

                index = fieldDescriptor.end()
                if (index < len(pathString)) and (pathString[index] == '/'):
                    # consume '/'
                    index += 1

            elif pathString[index] == '[':
                # consume '['
                index += 1

                dimensions = []
                dimensionCount = 0
                while dimensionCount < coda.CODA_MAX_NUM_DIMS:
                    dimensionDescriptor = dimensionRegEx.match(pathString, index)
                    if dimensionDescriptor is not None:
                        (idx,) = dimensionDescriptor.groups()

                        if idx is None:
                            raise PathError("invalid index")

                        idx = idx.strip()
                        if len(idx) == 0:
                            raise PathError("invalid index")

                        dimensions.append(int(idx))

                        index = dimensionDescriptor.end()
                        if index < len(pathString):
                            if pathString[index] == ',':
                                # consume ','
                                index += 1
                            elif pathString[index] == ']':
                                # consume ']'
                                index += 1
                                break
                            else:
                                raise PathError("invalid index")
                    else:
                        raise PathError("invalid index")

                    dimensionCount += 1

                # check for a period after the dimension specification
                if index < len(pathString):
                    if pathString[index] == '/':
                        # consume '/'
                        index += 1
                    elif pathString[index] != '[':
                        raise PathError("'/' or '[' expected")

                path.append(dimensions)
            else:
                raise PathError("extra characters in path")

        return path

    def GetSelectedNode(self):
        return self.selectedNode

    def notify(self, obj):
        if not self.productRootNode or self.autoExpanding:
            return

        pathString = obj.GetValue()
        try:
            path = self.ParsePath(pathString)
        except PathError as ex:
            message = wx.MessageDialog(self.frame, "Unable to parse path specification:\n%s" % (str(ex),),
                                       "Parse Error", style=wx.OK | wx.ICON_ERROR | wx.STAY_ON_TOP)
            message.ShowModal()
            return

        self.autoExpanding = True
        self.autoPath = path
        self.autoIndex = 0

        path.insert(0, self.productRootNode.name)
        self.view.ExpandPath(path)


class NavigationView(wx.ScrolledWindow):

    defaultTreeStyle = wx.TR_HAS_BUTTONS | wx.TR_NO_LINES | wx.NO_BORDER | wx.FULL_REPAINT_ON_RESIZE

    def __init__(self, parent, id, controller):
        wx.ScrolledWindow.__init__(self, parent, id, size=(100, 100), style=wx.NO_BORDER)

        # weakref to avoid garbage cycle.
        self.controller = weakref.proxy(controller)

        self.viewIndex = {}
        self.viewStack = []
        self.separatorStack = []

        self.sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.SetSizer(self.sizer)

        self.Bind(EVT_ITEM_SELECTED, self.OnItemSelected)
        self.Bind(EVT_CHILD_FOCUS, self.OnChildFocus)

        if sys.platform == "win32":
            self.Bind(EVT_ITEM_DESELECTED, self.OnItemDeselected)

    def ViewFactory(self, node):
        if node.type == model.TYPE_RECORD and not node.dirty() and len(node) > 0:
            view = RecordView(self, wx.NewId())
            view.Initialize(node)
            return view

        elif node.type == model.TYPE_ARRAY and node.isObjectArray() and len(node) > 0:
            if sys.platform == "win32":
                return NonVirtualIndexView(self, wx.NewId(), node)
            else:
                return IndexView(self, wx.NewId(), node)
        else:
            return None

    def PushView(self, view):
        if view is None:
            return

        self.viewIndex[view.GetId()] = len(self.viewStack)
        self.viewStack.append(view)

        # horizontal sizer
        self.sizer.Add(view, 0, wx.EXPAND, 0)
        separator = wx.StaticLine(self, -1, style=wx.LI_VERTICAL)
        self.separatorStack.append(separator)
        # horizontal sizer
        self.sizer.Add(separator, 0, wx.EXPAND | wx.LEFT | wx.RIGHT, 5)

    def PopView(self):
        # cannot pop the root view.
        assert len(self.viewStack) > 1, "Cannot pop root view"

        view = self.viewStack.pop()
        del self.viewIndex[view.GetId()]

        separator = self.separatorStack.pop()
        self.sizer.Detach(separator)
        separator.Destroy()

        self.sizer.Detach(view)
        view.Destroy()

    def PopUntil(self, view):
        while len(self.viewStack) > 0 and not self.viewStack[-1] == view:
            self.PopView()

    def ExpandPath(self, path):
        self.viewStack[0].SelectItem(0)

    def LayoutAndRefresh(self):
        self.Layout()
        # takes care of displaying horizontal scrollbar after
        # selection change...
        self.FitInside()

    def ScrollToEnd(self):
        vx, vy = self.GetVirtualSize()
        cx, cy = self.GetClientSize()
        xUnit, yUnit = self.GetScrollPixelsPerUnit()

        if vx > cx:
            scrollAmount = ((vx - cx) / xUnit) + 1
            self.Scroll(scrollAmount, -1)

    def ValidEvent(self, event):
        if (event.GetId() in self.viewIndex):
            return True
        else:
            return False

    def GetViewCount(self):
        return len(self.viewStack)

    def GetView(self, index):
        assert (index >= 0) and (index < len(self.viewStack)), "view index out of range"
        return self.viewStack[index]

    def GetViewIndex(self, view):
        try:
            return self.viewIndex[view.GetId()]
        except KeyError:
            return -1

    def OnItemDeselected(self, event):
        if self.ValidEvent(event):
            view = event.GetEventObject()
            self.controller.OnItemDeselected(view, view.GetNode(), event.index)
            self.LayoutAndRefresh()

    def OnItemSelected(self, event):
        if self.ValidEvent(event):
            view = event.GetEventObject()
            self.controller.OnItemSelected(view, view.GetNode(), event.index)

    def OnChildFocus(self, event):
        if self.ValidEvent(event):
            view = event.GetEventObject()
            self.controller.OnChildFocus(view, view.GetNode())


class RecordView(wx.ListCtrl):

    def __init__(self, parent, id):
        style = wx.LC_REPORT | wx.LC_NO_HEADER | wx.BORDER_NONE | wx.LC_SINGLE_SEL | wx.FULL_REPAINT_ON_RESIZE
        wx.ListCtrl.__init__(self, parent, id, style=style)

        self.SetImageList(model.iconImageList, wx.IMAGE_LIST_SMALL)
        self.InsertColumn(0, "Name")
        self.node = None
        self.index = -1

    def Initialize(self, node):
        assert not node.dirty(), "cannot create view for dirty record nodes"
        self.node = node

        for i in range(0, len(node)):
            field = node[i]
            if not (coda.get_option_filter_record_fields() and (field.hidden or not field.available)):
                self.InsertItem(model.GetNodeAsListItem(self.GetItemCount(), field, i))

        self.SetColumnWidth(0, wx.LIST_AUTOSIZE)
        self.SetMinSize(wx.Size(self.GetColumnWidth(0) + 25, 0))

        self.Bind(wx.EVT_LIST_ITEM_SELECTED, self.OnItemSelected)
        self.Bind(wx.EVT_LIST_ITEM_DESELECTED, self.OnItemDeselected)
        self.Bind(wx.EVT_SET_FOCUS, self.OnFocus)

    def GetNode(self):
        return self.node

    def GetSelectedIndex(self):
        if self.index >= 0:
            item = self.GetItem(self.index)
            return item.GetData()
        else:
            return -1

    def SelectItem(self, index):
        # programmatically selecting an item only causes an EVT_LIST_ITEM_SELECTED
        # event when the item is not currently selected.
        self.SetItemState(index, 0, wx.LIST_STATE_SELECTED | wx.LIST_STATE_FOCUSED)
        self.SetItemState(index, wx.LIST_STATE_SELECTED | wx.LIST_STATE_FOCUSED,
                          wx.LIST_STATE_SELECTED | wx.LIST_STATE_FOCUSED)

    def Find(self, destination):
        if not isinstance(destination, str):
            return -1
        else:
            return self.FindItem(-1, destination)

    def OnItemSelected(self, event):
        if event.GetIndex() != self.index:
            self.index = event.GetIndex()
            newEvent = ItemSelectedEvent(self.GetId(), event.GetItem().GetData())
            newEvent.SetEventObject(self)
            self.GetEventHandler().ProcessEvent(newEvent)

    def OnItemDeselected(self, event):
        self.index = -1
        if sys.platform == "win32":
            newEvent = ItemDeselectedEvent(self.GetId(), event.GetItem().GetData())
            newEvent.SetEventObject(self)
            self.GetEventHandler().ProcessEvent(newEvent)

    def OnFocus(self, event):
        newEvent = ChildFocusEvent(self.GetId())
        newEvent.SetEventObject(self)
        self.GetEventHandler().ProcessEvent(newEvent)
        event.Skip()


class NonVirtualIndexView(wx.ListCtrl):
    # This class was created because with the current version of wxPython
    # a virtual ListCtrl and a non-virtual ListCtrl differ in the events
    # they send and/or do not send on Windows platforms.
    #
    # On Windows, contrary to GTK, a ListCtrl with the LC_SINGLE_SEL style
    # allows the user to deselect the selected item by clicking under or
    # next to the list of items. On GTK, there always is a selected item.
    # When the user deselects the selected item (on Windows), then the native
    # control reflects this (the item is no longer highlighted). Therefore, some
    # reaction of the user interface required. Otherwise it would be possible to
    # deselect any node on the current path.
    #
    # On Windows, an EVT_LIST_ITEM_DESELECTED event is sent when an item is deselected.
    # NB. this implies the event is also sent when the user selects a different item,
    # in which case it is immediately followed by an EVT_LIST_ITEM_SELECTED event.
    #
    # The problem is that if the ListCtrl also has the LC_VIRTUAL style, _NO_ deselect
    # events are sent. So, the item looks deselected on screen, but there is no way for the
    # code to know this has happened! Therefore, on Windows a non-virtual version of the
    # IndexView class is used to prevent this situation (non-virtual controls _do_ sent
    # deselect events).
    #
    # If this issue is resolved in future versions of wxPython it would be wise to remove
    # this class and use the virtual IndexView class. A list of array indices may be very long
    # and a non-virtual ListCtrl has to add an item for each array index to the list in its
    # constructor. A virtual control will only request data for items that are visible,
    # through the OnGetItemText(), OnGetItemImage(), and OnGetItemAttr() methods. This is much
    # more efficient and therefore a virtual ListCtrl should be used whenever possible.
    #
    # NB. When removing this class, do not forget to check the whole event system. Look
    # for lines like 'if sys.platform ==/!= "win32"'.

    def __init__(self, parent, id, node):
        style = wx.LC_REPORT | wx.LC_NO_HEADER | wx.BORDER_NONE | wx.LC_SINGLE_SEL | wx.FULL_REPAINT_ON_RESIZE
        wx.ListCtrl.__init__(self, parent, id, style=style)
        assert len(node) > 0, "Number of elements should always be larger than zero"

        self.node = node

        if node.isRankZero():
            self.InsertColumn(0, "0", wx.LIST_FORMAT_RIGHT)
            self.InsertItem(0, "0")
            self.SetColumnWidth(0, wx.LIST_AUTOSIZE)
            self.SetSizeHints(self.GetColumnWidth(0) + 25, 0)
        else:
            rank = len(node.dimensions)

            # add columns.
            for i in range(0, rank):
                self.InsertColumn(i, str(i), wx.LIST_FORMAT_RIGHT)

            # add rows.
            multiDimensionalIndex = [0] * rank
            for row in range(0, len(node)):
                if rank == 1:
                    self.InsertItem(row, str(multiDimensionalIndex[0]))
                else:
                    item = self.InsertItem(row, str(multiDimensionalIndex[0]) + ",")

                    # insert item into list
                    for col in range(1, rank - 1):
                        self.SetItem(item, col, str(multiDimensionalIndex[col]) + ",")
                    self.SetItem(item, rank - 1, str(multiDimensionalIndex[-1]))

                # update dimensions
                for col in reversed(list(range(0, rank))):
                    multiDimensionalIndex[col] += 1
                    if multiDimensionalIndex[col] < node.dimensions[col]:
                        break
                    multiDimensionalIndex[col] = 0

            # compute total width
            width = 0
            for i in range(0, rank):
                self.SetColumnWidth(i, wx.LIST_AUTOSIZE)
                width += self.GetColumnWidth(i)
            self.SetSizeHints(width + 25, 0)

        self.index = -1

        self.Bind(wx.EVT_LIST_ITEM_SELECTED, self.OnItemSelected)
        self.Bind(wx.EVT_LIST_ITEM_DESELECTED, self.OnItemDeselected)
        self.Bind(wx.EVT_SET_FOCUS, self.OnFocus)

    def GetNode(self):
        return self.node

    def GetSelectedIndex(self):
        return self.index

    def SelectItem(self, index):
        # programmatically selecting an item only causes an EVT_LIST_ITEM_SELECTED
        # event when the item is not currently selected.
        self.SetItemState(index, 0, wx.LIST_STATE_SELECTED | wx.LIST_STATE_FOCUSED)
        self.SetItemState(index, wx.LIST_STATE_SELECTED | wx.LIST_STATE_FOCUSED,
                          wx.LIST_STATE_SELECTED | wx.LIST_STATE_FOCUSED)

    def Find(self, destination):
        if not isinstance(destination, list):
            return -1

        if self.node.isRankZero():
            if len(destination) != 1:
                return -1

            index = destination[0]
        else:
            if len(destination) != len(self.node.dimensions):
                return -1

            # compute flat index via Horner's rule.
            index = 0
            for i in range(1, len(self.node.dimensions)):
                index = (index + destination[i - 1]) * self.node.dimensions[i]
            index += destination[-1]

        if (index < 0) or (index > self.GetItemCount() - 1):
            return -1
        else:
            return index

    def OnItemSelected(self, event):
        if event.GetIndex() != self.index:
            self.index = event.GetIndex()
            newEvent = ItemSelectedEvent(self.GetId(), self.index)
            newEvent.SetEventObject(self)
            self.GetEventHandler().ProcessEvent(newEvent)

    def OnItemDeselected(self, event):
        self.index = -1
        if sys.platform == "win32":
            newEvent = ItemDeselectedEvent(self.GetId(), event.GetIndex())
            newEvent.SetEventObject(self)
            self.GetEventHandler().ProcessEvent(newEvent)

    def OnFocus(self, event):
        newEvent = ChildFocusEvent(self.GetId())
        newEvent.SetEventObject(self)
        self.GetEventHandler().ProcessEvent(newEvent)
        event.Skip()


class IndexView(wx.ListCtrl):

    def __init__(self, parent, id, node):
        style = wx.LC_REPORT | wx.LC_NO_HEADER | wx.BORDER_NONE | wx.LC_VIRTUAL | wx.LC_SINGLE_SEL | \
            wx.FULL_REPAINT_ON_RESIZE
        wx.ListCtrl.__init__(self, parent, id, style=style)

        assert len(node) > 0, "Number of elements should always be larger than zero"

        self.node = node

        if node.isRankZero():
            self.InsertColumn(0, "0", wx.LIST_FORMAT_RIGHT)

            # Alas! GTK1.2's implementation of GetTextExtent() does not
            # seem to be very accurate.
            if sys.platform != "win32":
                columnString = "9" * (1 + 2)
            else:
                columnString = "9"
            width, height = self.GetTextExtent(columnString)

            self.SetColumnWidth(0, width)
            self.SetSizeHints(width + 25, 0)
        else:
            width = 0
            for i in range(0, len(node.dimensions)):
                self.InsertColumn(i, str(i), wx.LIST_FORMAT_RIGHT)

                if node.dimensions[i] == 1:
                    digitCount = 1
                else:
                    digitCount = int(math.log10(node.dimensions[i] - 1)) + 1

                # Alas! GTK1.2's implementation of GetTextExtent() does not
                # seem to be very accurate.
                if sys.platform != "win32":
                    digitCount += 2

                columnString = "9" * digitCount
                if i < len(node.dimensions) - 1:
                    columnString += ","
                columnWidth, rowHeight = self.GetTextExtent(columnString)

                self.SetColumnWidth(i, columnWidth)
                width += columnWidth

            self.SetSizeHints(width + 25, 0)

        self.SetItemCount(len(node))

        self.index = -1

        self.Bind(wx.EVT_LIST_ITEM_SELECTED, self.OnItemSelected)
        self.Bind(wx.EVT_LIST_ITEM_DESELECTED, self.OnItemDeselected)
        self.Bind(wx.EVT_SET_FOCUS, self.OnFocus)

    def GetNode(self):
        return self.node

    def GetSelectedIndex(self):
        return self.index

    def SelectItem(self, index):
        # programmatically selecting an item only causes an EVT_LIST_ITEM_SELECTED
        # event when the item is not currently selected.
        self.SetItemState(index, 0, wx.LIST_STATE_SELECTED | wx.LIST_STATE_FOCUSED)
        self.SetItemState(index, wx.LIST_STATE_SELECTED | wx.LIST_STATE_FOCUSED,
                          wx.LIST_STATE_SELECTED | wx.LIST_STATE_FOCUSED)

    def Find(self, destination):
        if not isinstance(destination, list):
            return -1

        if self.node.isRankZero():
            if len(destination) != 1:
                return -1

            index = destination[0]
        else:
            if len(destination) != len(self.node.dimensions):
                return -1

            # compute flat index via Horner's rule.
            index = 0
            for i in range(1, len(self.node.dimensions)):
                index = (index + destination[i - 1]) * self.node.dimensions[i]
            index += destination[-1]

        if (index < 0) or (index > self.GetItemCount() - 1):
            return -1
        else:
            return index

    def OnItemSelected(self, event):
        if event.GetIndex() != self.index:
            self.index = event.GetIndex()
            newEvent = ItemSelectedEvent(self.GetId(), self.index)
            newEvent.SetEventObject(self)
            self.GetEventHandler().ProcessEvent(newEvent)

    def OnItemDeselected(self, event):
        self.index = -1
        if sys.platform == "win32":
            newEvent = ItemDeselectedEvent(self.GetId(), event.GetIndex())
            newEvent.SetEventObject(self)
            self.GetEventHandler().ProcessEvent(newEvent)

    def OnFocus(self, event):
        newEvent = ChildFocusEvent(self.GetId())
        newEvent.SetEventObject(self)
        self.GetEventHandler().ProcessEvent(newEvent)
        event.Skip()

    def OnGetItemText(self, item, col):
        multiDimensionalIndex = model.GetMultiDimensionalIndex(self.node, item)

        if self.node.isRankZero():
            return str(multiDimensionalIndex[0])
        elif col < len(self.node.dimensions) - 1:
            return str(multiDimensionalIndex[col]) + ","
        else:
            return str(multiDimensionalIndex[col])

    def OnGetItemImage(self, item):
        return -1

    def OnGetItemAttr(self, item):
        return None


class AttributeView(wx.ListCtrl):

    def __init__(self, parent, id):
        style = wx.LC_REPORT | wx.BORDER_SUNKEN | wx.LC_VRULES | wx.LC_SINGLE_SEL | wx.FULL_REPAINT_ON_RESIZE
        wx.ListCtrl.__init__(self, parent, id, style=style)

        self.InsertColumn(0, "Name")
        self.InsertColumn(1, "Value")

    def Initialize(self, frame, node):
        self._AppendItem(("Type", model.GetTypeAsString(node.type)))

        if node.type == model.TYPE_ARRAY:
            self._AppendItem(("Elements", "%u" % (len(node),)))

            if node.isRankZero():
                self._AppendItem(("Dimensions", ""))
            else:
                dimensionString = ""
                for dim in node.dimensions[:-1]:
                    dimensionString += "%u x " % (dim,)
                dimensionString += "%u" % (node.dimensions[-1],)
                self._AppendItem(("Dimensions", dimensionString))

            self._AppendItem(("Base type", model.GetTypeAsString(node.base)))
        elif node.type == model.TYPE_RECORD:
            self._AppendItem(("Fields", "%u" % (len(node),)))
        elif node.type == model.TYPE_STRING:
            self._AppendItem(("Length", "%u" % (len(node),)))

        # save the number of succesfully inserted attributes
        safeCount = self.GetItemCount()

        try:
            if wx.Config.Get().ReadBool('ProductBrowser/ShowByteSize', False):
                # size in bytes:bits.
                bitSize = node.getBitSize()
                if bitSize != -1:
                    bitFraction = bitSize & 7
                    if bitFraction == 0:
                        self._AppendItem(("Size", "%u" % (bitSize / 8,)))
                    else:
                        self._AppendItem(("Size", "%u:%u" % (bitSize / 8, bitFraction)))

            # get node type.
            nodeCodaType = coda.cursor_get_type(node.cursor)

            # description.
            description = coda.type_get_description(nodeCodaType)
            if description is not None:
                self._AppendItem(("Description", description))

            # unit of measurement.
            unit = coda.type_get_unit(nodeCodaType)
            if unit is not None:
                self._AppendItem(("Unit", unit))

            # check for fixed value.
            nodeCodaClass = coda.type_get_class(nodeCodaType)
            if (nodeCodaClass == coda.coda_text_class) or (nodeCodaClass == coda.coda_raw_class):
                type = model.TranslateCodaType(nodeCodaClass, nodeCodaType)
                fixedValue = coda.type_get_fixed_value(nodeCodaType)
                if fixedValue is not None:
                    # a maximum size of _maxLineLength characters is specified, because
                    # GDK may crash if the length of a string item gets too long.
                    # also, if the line is really long the user will probably not want
                    # to view it as a single line in a ListCtrl anyway.
                    self._AppendItem(("Fixed value", model.GetDataAsString(type, fixedValue, _maxLineLength)))

        except coda.CodacError as ex:
            for i in range(safeCount, self.GetItemCount()):
                self.DeleteItem(i)

            message = wx.MessageDialog(frame, "Error retrieving attribute information.\n\n"
                                       "CODA error message: \"%s\"" % str(ex), "CODA error",
                                       style=wx.OK | wx.ICON_ERROR | wx.STAY_ON_TOP)
            message.ShowModal()

        # auto-size the list
        self._AutoSize((0, 1))

    def _AppendItem(self, descriptor):
        index = self.InsertItem(self.GetItemCount(), descriptor[0])
        self.SetItem(index, 1, descriptor[1])

    def _AutoSize(self, columns):
        for col in columns:
            header = self.GetColumn(col)
            if header:
                self.SetColumnWidth(col, wx.LIST_AUTOSIZE)
                columnWidth = self.GetColumnWidth(col)
                headerWidth, headerHeight = self.GetTextExtent(header.GetText())
                self.SetColumnWidth(col, max(columnWidth, headerWidth + 15))


class ProductAttributeView(wx.ListCtrl):

    def __init__(self, parent, id):
        style = wx.LC_REPORT | wx.BORDER_SUNKEN | wx.LC_VRULES | wx.LC_SINGLE_SEL | wx.FULL_REPAINT_ON_RESIZE
        wx.ListCtrl.__init__(self, parent, id, style=style)

        self.InsertColumn(0, "Fieldname")
        self.InsertColumn(1, "Real name")
        self.InsertColumn(2, "Value")
        self.InsertColumn(3, "Type")
        self.InsertColumn(4, "Array Dimensions")

    def Initialize(self, frame, attributes):
        try:
            data = coda.fetch(attributes.cursor)
        except (coda.CodaError, coda.CodacError) as ex:
            message = wx.MessageDialog(frame, "Error reading attribute information from product.\nPlease check product "
                                       "sanity, e.g. using codacheck.\n\nCODA error message: \"%s\"" % str(ex),
                                       "CODA error", style=wx.OK | wx.ICON_ERROR | wx.STAY_ON_TOP)
            message.ShowModal()
        else:
            for i in range(0, len(attributes)):
                attr = attributes[i]

                if attr.available:
                    if attr.type == model.TYPE_ARRAY:
                        valueString = ""
                        if data[i].size > 0:
                            value = data[i].flat
                            for val in value[:-1]:
                                valueString += "%s, " % (model.GetDataAsString(attr.base, val),)
                            valueString += "%s" % (model.GetDataAsString(attr.base, value[-1]),)

                        if len(valueString) > _maxLineLength:
                            valueString = "<array>"

                        dimensionString = ""
                        if not attr.isRankZero():
                            dimensionString = ""
                            for dim in attr.dimensions[:-1]:
                                dimensionString += "%u x " % (dim,)
                            dimensionString += "%u" % (attr.dimensions[-1],)

                        index = self.InsertItem(i, attr.name)
                        self.SetItem(index, 1, attr.real_name)
                        self.SetItem(index, 2, valueString)
                        self.SetItem(index, 3, model.GetTypeAsString(attr.base))
                        self.SetItem(index, 4, dimensionString)
                    else:
                        index = self.InsertItem(i, attr.name)
                        # a maximum size of _maxLineLength characters is specified, because
                        # GDK may crash if the length of a string item gets too long.
                        # also, if the line is really long the user will probably not want
                        # to view as a single line in a ListCtrl anyway.
                        self.SetItem(index, 1, attr.real_name)
                        self.SetItem(index, 2, model.GetDataAsString(attr.type, data[i], _maxLineLength))
                        self.SetItem(index, 3, model.GetTypeAsString(attr.type))
                        self.SetItem(index, 4, "-")
                else:
                    index = self.InsertItem(i, attr.name)
                    self.SetItemTextColour(index, "light grey")

        self.SetColumnWidth(2, 150)
        self._AutoSize((0, 1, 3, 4))

    def _AutoSize(self, columns):
        for col in columns:
            header = self.GetColumn(col)
            if header:
                self.SetColumnWidth(col, wx.LIST_AUTOSIZE)
                columnWidth = self.GetColumnWidth(col)
                headerWidth, headerHeight = self.GetTextExtent(header.GetText())
                self.SetColumnWidth(col, max(columnWidth, headerWidth + 15))


class NodeAttributeView(wx.Panel):

    def __init__(self, parent, id):
        wx.Panel.__init__(self, parent, id)
        # this corrects the colour of the panel in win32 (without this
        # line it defaults to dark grey, instead of light grey).
        self.SetBackgroundColour(self.GetBackgroundColour())

    def Initialize(self, frame, node):
        sizer = wx.BoxSizer(wx.VERTICAL)

        view = AttributeView(self, -1)
        view.Initialize(frame, node)
        sizer.Add(wx.StaticText(self, -1, "Attributes from Product Format Definition"), 0, wx.ALIGN_CENTRE | wx.ALL, 5)
        sizer.Add(wx.StaticLine(self, -1, style=wx.LI_HORIZONTAL), 0, wx.EXPAND, 0)
        sizer.Add(view, 1, wx.EXPAND | wx.ALL, 5)

        try:
            attributes = model.RetrieveAttributes(node)
            model.RetrieveFieldInfo(attributes)
        except (coda.CodaError, coda.CodacError) as ex:
            message = wx.MessageDialog(frame, "Error reading attribute information from product.\nPlease check product "
                                       "sanity, e.g. using codacheck.\n\nCODA error message: \"%s\"" % str(ex),
                                       "CODA error", style=wx.OK | wx.ICON_ERROR | wx.STAY_ON_TOP)
            message.ShowModal()
        else:
            if len(attributes) > 0:
                view = ProductAttributeView(self, -1)
                view.Initialize(frame, attributes)
                sizer.Add(wx.StaticText(self, -1, "Attributes from Product"), 0, wx.ALIGN_CENTRE | wx.ALL, 5)
                sizer.Add(wx.StaticLine(self, -1, style=wx.LI_HORIZONTAL), 0, wx.EXPAND, 0)
                sizer.Add(view, 2, wx.EXPAND | wx.ALL, 5)

        self.SetSizer(sizer)
        self.Layout()


class NodeDataView(wx.Panel):
    # a mapping from type to renderers

    renderers = {}

    def RegisterRenderer(cls, type, renderer):
        try:
            list = cls.renderers[type]
        except KeyError:
            list = []
            cls.renderers[type] = list

        if renderer not in list:
            list.append(renderer)

    RegisterRenderer = classmethod(RegisterRenderer)

    def __init__(self, parent, id, frame, node):
        wx.Panel.__init__(self, parent, id)
        # this corrects the colour of the panel in win32 (without this
        # line it defaults to dark grey, instead of light grey).
        self.SetBackgroundColour(self.GetBackgroundColour())

        self.frame = weakref.ref(frame)
        self.node = node

        # get a list of renderers capable of rendering
        # the specified node.
        try:
            availableRenderers = NodeDataView.renderers[node.type]

            # purge renderers from the list that cannot render
            # the specific node given. for instance, a PlotRenderer
            # is able to render nodes of TYPE_ARRAY but only with
            # a specific base type (e.g. uint32, but not string or char).
            self.qualifiedRenderers = [x for x in availableRenderers if x.CanRender(node)]

        except KeyError:
            self.qualifiedRenderers = []

        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.sizer.Add((5, 5), 0, 0)

        if len(self.qualifiedRenderers) == 0:
            view = renderers.UnavailableRenderer(self)
        else:
            view = self.qualifiedRenderers[0](self)
            try:
                view.Initialize(self.node)
            except renderers.RenderError as ex:
                view.Destroy()
                view = renderers.UnavailableRenderer(self)

                message = wx.MessageDialog(self.frame(), "Could not instantiate renderer.\nPlease check product "
                                           "sanity, e.g. using codacheck.\n\nError message: \"%s\"" % (str(ex),),
                                           "Render error", style=wx.OK | wx.ICON_ERROR | wx.STAY_ON_TOP)
                message.ShowModal()

            # if multiple renderers are qualified, create a toggle
            # button for each renderer and cache the views.
            if len(self.qualifiedRenderers) > 1:
                self.buttonGroup = ButtonGroup()
                buttonSizer = wx.BoxSizer(wx.HORIZONTAL)
                buttonSizer.Add(wx.StaticText(self, -1, "View data as:"), 0, wx.ALIGN_CENTRE | wx.ALL, 5)
                for renderer in self.qualifiedRenderers:
                    button = wx.ToggleButton(self, -1, renderer.name)
                    buttonSizer.Add(button, 0, wx.ALL, 5)
                    self.buttonGroup.AppendButton(button)

                self.sizer.Add(buttonSizer, 0, wx.EXPAND, 0)
                self.sizer.Add(wx.StaticLine(self, -1, style=wx.LI_HORIZONTAL), 0, wx.EXPAND | wx.LEFT | wx.RIGHT, 5)
                self.sizer.Add((5, 5), 0, 0)

                self.views = {}
                self.views[0] = view

                self.Bind(wx.EVT_TOGGLEBUTTON, self.OnToggleButton)

        # add view to sizer
        self.sizer.Add(view, 1, wx.EXPAND, 0)
        self.SetSizer(self.sizer)
        self.Layout()

    def OnToggleButton(self, event):
        currentView = self.buttonGroup.GetSelected()
        if self.buttonGroup.ProcessToggleEvent(event.GetEventObject()):
            nextView = self.buttonGroup.GetSelected()

            try:
                view = self.views[nextView]
            except KeyError:
                # view not in cache, so create it.
                view = self.qualifiedRenderers[nextView](self)

                try:
                    view.Initialize(self.node)

                except renderers.RenderError as ex:
                    view.Destroy()
                    view = renderers.UnavailableRenderer(self)

                    message = wx.MessageDialog(self.frame(), "Could not instantiate renderer.\nPlease check product "
                                               "sanity, e.g. using codacheck.\n\nError message: \"%s\"" % (str(ex),),
                                               "Render error", style=wx.OK | wx.ICON_ERROR | wx.STAY_ON_TOP)
                    message.ShowModal()

                # add view to cache.
                self.views[nextView] = view
                self.sizer.Add(view, 1, wx.EXPAND, 0)

            # hide current view
            self.sizer.Show(self.views[currentView], False, True)

            # show next view, if creation was succesful.
            self.sizer.Show(view, True, True)
            self.sizer.Layout()


class NodeView(wx.Notebook):

    def __init__(self, parent, id, frame):
        wx.Notebook.__init__(self, parent, id, style=wx.CLIP_CHILDREN | wx.NB_BOTTOM)
        self.frame = weakref.ref(frame)

    def notify(self, obj):
        self.DeleteAllPages()
        if obj.GetSelectedNode() is not None:
            view = NodeDataView(self, -1, self.frame(), obj.GetSelectedNode())
            self.AddPage(view, "Data", True)

            view = NodeAttributeView(self, -1)
            view.Initialize(self.frame(), obj.GetSelectedNode())
            self.AddPage(view, "Attributes", False)


class PathView(wx.TextCtrl, Observable):

    def __init__(self, parent):
        wx.TextCtrl.__init__(self, parent, -1, "", style=wx.TE_PROCESS_ENTER)
        Observable.__init__(self)

        self.Bind(wx.EVT_TEXT_ENTER, self.OnTextEnter)

    def OnTextEnter(self, event):
        wx.CallAfter(self.notifyObservers)
        event.Skip()

    def notify(self, obj):
        if obj.GetSelectedNode() is not None:
            path = obj.GetSelectedNode().path()
            # strip rootNode and productRootNode
            self.SetValue(model.GetPathAsString(path[2:]))
        else:
            self.SetValue("")


class PathDropTarget(wx.TextDropTarget, Observable):

    def __init__(self):
        wx.TextDropTarget.__init__(self)
        Observable.__init__(self)
        self.path = None

    def OnDropText(self, x, y, data):
        self.path = data
        wx.CallAfter(self.notifyObservers)

    def GetValue(self):
        return self.path


class ProductBrowser(wx.Frame):

    initialised = False

    def __init__(self, parent, filename):
        wx.Frame.__init__(self, parent, -1, os.path.basename(str(filename)), size=(800, 600))

        if not ProductBrowser.initialised:
            model.InitIcons()
            ProductBrowser.initialised = True

        if 8 * struct.calcsize("P") == 32:
            # On a 32-bit platform we open a product without using mmap.
            # This allows us to have multiple large product files open without
            # having the risk of running out of memory address space.
            prev_opt = coda.get_option_use_mmap()
            coda.set_option_use_mmap(0)
            try:
                self.product = coda.open(str(filename))
            finally:
                coda.set_option_use_mmap(prev_opt)
        else:
            self.product = coda.open(str(filename))

        try:
            self.cursor = coda.Cursor()
            coda.cursor_set_product(self.cursor, self.product)

            self.closing = False

            # register node renderers
            self.RegisterRenderers()

            # create GUI
            self.CreateMenuBar()
            self.CreateControls(self.product, self.cursor)

            # Other Event Listeneres
            self.Bind(wx.EVT_CLOSE, self.OnClose)
        except Exception:
            coda.close(self.product)
            raise

    def CreateMenuBar(self):
        menuBar = wx.MenuBar()

        filemenu = wx.Menu()

        item = filemenu.Append(wx.ID_CLOSE, "&Close\tCtrl+W", "Close the Product Browser")
        self.Bind(wx.EVT_MENU, self.OnClose, item)

        menuBar.Append(filemenu, "&File")

        self.SetMenuBar(menuBar)

    def CreateControls(self, product, cursor):
        # this panel seems to be required for win32, because the default
        # background color of a frame is dark grey, and the trick to correct
        # the default background color of a panel (see below) does not seem
        # to work for frames.
        panel = wx.Panel(self, -1, style=wx.NO_BORDER)

        # this corrects the colour of the panel in win32 (without this
        # line it defaults to dark grey, instead of light grey).
        panel.SetBackgroundColour(panel.GetBackgroundColour())

        self.pathView = PathView(panel)

        backButton = wx.BitmapButton(panel, -1, wx.ArtProvider.GetBitmap(wx.ART_GO_BACK, wx.ART_OTHER, (16, 16)),
                                     size=(23, 23))
        backButton.Bind(wx.EVT_BUTTON, self.OnBack)

        forwardButton = wx.BitmapButton(panel, -1, wx.ArtProvider.GetBitmap(wx.ART_GO_FORWARD, wx.ART_OTHER, (16, 16)),
                                        size=(23, 23))
        forwardButton.Bind(wx.EVT_BUTTON, self.OnForward)

        copyCmdButton = wx.BitmapButton(panel, -1, wx.ArtProvider.GetBitmap(wx.ART_COPY, wx.ART_OTHER, (16, 16)),
                                        size=(23, 23))
        copyCmdButton.Bind(wx.EVT_BUTTON, self.OnCopyCmd)

        pathSizer = wx.BoxSizer(wx.HORIZONTAL)
        pathSizer.Add(backButton, 0, wx.ALIGN_CENTER | wx.LEFT | wx.RIGHT, 5)
        pathSizer.Add(forwardButton, 0, wx.ALIGN_CENTER | wx.RIGHT, 5)
        pathSizer.Add(self.pathView, 1, wx.RIGHT, 5)
        pathSizer.Add(copyCmdButton, 0, wx.ALIGN_CENTER | wx.RIGHT, 5)

        splitterstyle = wx.SP_LIVE_UPDATE | wx.SIMPLE_BORDER
        if wx.Platform == "__WXMAC__":
            splitterstyle |= wx.SP_3DSASH
        splitter = wx.SplitterWindow(panel, -1, style=splitterstyle)
        # this corrects the colour of the splitter in win32 (without this
        # line it defaults to dark grey, instead of light grey).
        splitter.SetBackgroundColour(splitter.GetBackgroundColour())
        splitter.SetMinimumPaneSize(50)

        self.navigationController = NavigationController(self)
        self.navigationController.register(self.pathView)
        self.pathView.register(self.navigationController)
        self.pathDropTarget = PathDropTarget()
        self.pathDropTarget.register(self.navigationController)
        self.SetDropTarget(self.pathDropTarget)

        navigationView = NavigationView(splitter, -1, self.navigationController)
        navigationView.SetScrollRate(5, 0)
        navigationView.SetBackgroundColour("white")

        self.nodeView = NodeView(splitter, -1, self)
        self.navigationController.register(self.nodeView)

        # the position of this line of code is important: Initialize() will cause
        # a call to notifyObservers(). therefore, all observers that want to be
        # notified of this initial event should register before this line.
        self.navigationController.Initialize(navigationView, cursor)

        splitter.SplitHorizontally(navigationView, self.nodeView, -300)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(pathSizer, 0, wx.EXPAND | wx.TOP | wx.BOTTOM, 5)
        sizer.Add(wx.StaticLine(panel, -1, style=wx.LI_HORIZONTAL), 0, wx.EXPAND | wx.BOTTOM, 3)
        sizer.Add(splitter, 1, wx.EXPAND, 0)

        panel.SetSizer(sizer)
        self.SetSizeHints(400, 300)

    def RegisterRenderers(self):
        NodeDataView.RegisterRenderer(model.TYPE_RECORD, renderers.RecordRenderer)
        NodeDataView.RegisterRenderer(model.TYPE_ARRAY, renderers.ArrayRenderer)
        NodeDataView.RegisterRenderer(model.TYPE_ARRAY, renderers.PlotRenderer)
        NodeDataView.RegisterRenderer(model.TYPE_INT8, renderers.ScalarRenderer)
        NodeDataView.RegisterRenderer(model.TYPE_UINT8, renderers.ScalarRenderer)
        NodeDataView.RegisterRenderer(model.TYPE_INT16, renderers.ScalarRenderer)
        NodeDataView.RegisterRenderer(model.TYPE_UINT16, renderers.ScalarRenderer)
        NodeDataView.RegisterRenderer(model.TYPE_INT32, renderers.ScalarRenderer)
        NodeDataView.RegisterRenderer(model.TYPE_UINT32, renderers.ScalarRenderer)
        NodeDataView.RegisterRenderer(model.TYPE_INT64, renderers.ScalarRenderer)
        NodeDataView.RegisterRenderer(model.TYPE_UINT64, renderers.ScalarRenderer)
        NodeDataView.RegisterRenderer(model.TYPE_VSF_INTEGER, renderers.ScalarRenderer)
        NodeDataView.RegisterRenderer(model.TYPE_FLOAT32, renderers.ScalarRenderer)
        NodeDataView.RegisterRenderer(model.TYPE_FLOAT64, renderers.ScalarRenderer)
        NodeDataView.RegisterRenderer(model.TYPE_CHAR, renderers.ScalarRenderer)
        NodeDataView.RegisterRenderer(model.TYPE_STRING, renderers.ScalarRenderer)
        NodeDataView.RegisterRenderer(model.TYPE_BYTES, renderers.BytesRenderer)
        NodeDataView.RegisterRenderer(model.TYPE_TIME, renderers.ScalarRenderer)
        NodeDataView.RegisterRenderer(model.TYPE_COMPLEX, renderers.ScalarRenderer)

    def OnBack(self, event):
        self.navigationController.Back()

    def OnForward(self, event):
        self.navigationController.Forward()

    def OnCopyCmd(self, event):
        selectedNode = self.navigationController.GetSelectedNode()
        if selectedNode:
            path = selectedNode.path()

            pathSpecification = ""
            for node in path[2:]:
                if isinstance(node.name, list):
                    if len(node.name) == 1:
                        pathSpecification += str(node.name[0])
                    else:
                        pathSpecification += str(node.name)
                else:
                    pathSpecification += "\"%s\"" % (node.name,)

                pathSpecification += ", "

            if wx.TheClipboard.Open():
                wx.TheClipboard.SetData(wx.TextDataObject(pathSpecification[:-2]))
                wx.TheClipboard.Close()

    def OnClose(self, event):
        if self.closing:
            return

        self.closing = True

        # unregister observers
        self.navigationController.unregister(self.pathView)
        self.pathView.unregister(self.navigationController)
        self.pathDropTarget.unregister(self.navigationController)
        self.navigationController.unregister(self.nodeView)

        # close product
        coda.close(self.product)

        if self.GetParent():
            self.GetParent().SetFocus()
        self.Destroy()



# Copyright (C) 2002-2022 S[&]T, The Netherlands.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import math
import numpy


class Pong:
    """ This is what happens if you get over-enthousiastic... [eddy]"""

    def __init__(self):
        self.nr_frames = 100

        self.bat1_x = [.005, .050, .050, .005, .005, .050, .005, .050, .005, .050]
        self.bat2_x = [.950, .995, .995, .950, .950, .995, .950, .995, .950, .995]

        self.bat_base_y = [0, 0, .5, .5, 0, .1, .2, .3, .4, .5]

        self.ball_base_x = .06
        self.ball_base_y = .5

        self.ball_state_x = 0
        self.ball_state_y = 0

        self.table_x = [0, 1, 1, 0, 0]
        self.table_y = [1, 1, 0, 0, 1]

        self.CalculateData()

    def bat1_func_y(self, y, x):
        return numpy.array(self.bat_base_y) + (numpy.sin((y + 1) * 2 * math.pi / self.nr_frames) / 2 + .5) / 2

    def bat2_func_y(self, y, x):
        return numpy.array(self.bat_base_y) + (numpy.cos((y + 10) * 4 * math.pi / self.nr_frames) / 2 + .5) / 2

    def ball_func_x(self, y, x):
        return 1 - (numpy.fabs((y - 50) / 56.8) + .06)

    def ball_func_y(self, y, x):
        return numpy.where(y < 25, ((y / 50.5) + .505) % 1, numpy.fabs(.995 - (y - 25) / 50.5))

    def CalculateData(self):
        import wx
        from visan.plot import PlotFrame

        plot = PlotFrame(wx.GetApp().GetTopWindow(), -1)
        dataSetId = plot.AddDataSet(self.table_x, self.table_y)
        plot.SetPlotColor(dataSetId, (0, 0, 0))
        plot.SetDataSetLabel(dataSetId, "Ping Pong Table")

        dataSetId = plot.AddDataSet(numpy.array(self.bat1_x),
                                    numpy.fromfunction(self.bat1_func_y, (self.nr_frames, len(self.bat_base_y))))
        plot.SetPlotColor(dataSetId, (0, 1.0, 0))
        plot.SetDataSetLabel(dataSetId, "Player 1")

        dataSetId = plot.AddDataSet(numpy.array(self.bat2_x),
                                    numpy.fromfunction(self.bat2_func_y, (self.nr_frames, len(self.bat_base_y))))
        plot.SetPlotColor(dataSetId, (0, 0, 1.0))
        plot.SetDataSetLabel(dataSetId, "Player 2")

        dataSetId = plot.AddDataSet(numpy.fromfunction(self.ball_func_x, (self.nr_frames, 1)),
                                    numpy.fromfunction(self.ball_func_y, (self.nr_frames, 1)))
        plot.SetPlotPoints(dataSetId, True)
        plot.SetPlotColor(dataSetId, (1.0, 0, 0))
        plot.SetPointSize(dataSetId, 3)
        plot.SetDataSetLabel(dataSetId, "The Ball")

        plot.SetPlotTitle("Pong")
        plot.SetXAxisRange((-0.1, 1.1))
        plot.SetYAxisRange((-0.1, 1.1))
        plot.ShowAnimationToolbar(False)

        plot.animationToolbar.SetLoop(True)
        plot.animationToolbar.SetFPS(40)
        plot.animationToolbar.Play()


def pong():
    Pong()



# This is an example VISAN script showing the extraction and plotting of ECMWF MARS data in GRIB format

# Make sure to set the 'products-file directory' option in the VISAN Preferences panel
# to a directory containing ECMWF MARS GRIB products.

# This example will then take the first product it finds in this directory and
# for that product plot the O3 column number density


def run():

    import glob
    import wx

    productdir = str(wx.Config.Get().Read('DirectoryLocation/Products'))

    # Use glob to find all files in productdir starting with 'C1D'
    files = glob.glob(os.path.join(productdir, "*.grib"))
    if len(files) == 0:
        print(("Could not find any GRIB files in directory '" + productdir + "'"))
        return

    # We only take the first file
    filename = files[0]

    # O3 is by default provided as mass column densities
    # During import we derive a new variable from this that contains a column number density with unit DU
    # And at the end we remove the mass column density variable again since we don't need it anymore
    data = harp.import_product(filename, "keep(latitude,longitude,O3_column_density);"
                               "derive(O3_column_number_density {latitude,longitude} [DU]);exclude(O3_column_density)")
    wplot(data, colorrange=[150, 500])


run()



# Copyright (C) 2002-2022 S[&]T, The Netherlands.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
VISAN command line functions and procedures
"""

__all__ = ['histogramplot', 'plot', 'wplot', 'executescript', 'browseproduct', 'version']

import os
import sys
import numpy
import harp
from .harpplot import PlotDataForProduct, WorldPlotDataForProduct


def histogramplot(data, bins, *args, **kwargs):
    """ Draw a histogram plot.

    This is a convenience function that creates a histogram in a plot window.

    >>> histogramplot(data, bins)

    will create a histogramplot where 'data' is the output of the histogram()
    function and 'bins' is the bins parameter that was used to create the
    histogram.

    Both 'data' and 'bins' should be rank-1 arrays and the array length of
    'data' should be one less than the length of 'bins'.

    For customizing the plot you can use the same additional options as are
    available for the plot() function.
    """

    data = numpy.asarray(data)
    bins = numpy.asarray(bins)
    x = numpy.transpose(numpy.reshape(numpy.concatenate([bins[:-1], bins[1:]]), (2, len(bins) - 1))).ravel()
    y = numpy.transpose(numpy.reshape(numpy.concatenate([data, data]), (2, len(data)))).ravel()
    plot(x, y, *args, **kwargs)


def plot(*args, **kwargs):
    """ Display a 2-dimensional plot of a dataset.

    The plot() function has several different usage modes, and
    can take a large number of optional keywords. For a full
    reference guide, please consult the VISAN documentation at:

        <http://www.stcorp.nl/beat/>

    Basic usage:

    >>> plot(ydata)

    where 'ydata' is a one-dimensional Python list or numpy array,
    will plot the 'ydata' values on the vertical axis against an
    index array [0..len(data)] on the horizontal axis, in a
    single plot frame.

    >>> plot(xdata, ydata)

    where 'xdata' and 'ydata' are both one-dimensional, will plot
    the 'ydata' values on the vertical axis against the
    corresponding 'xdata' values on the horizontal axis.

    If 'ydata', or 'xdata' and 'ydata' are two-dimensional n x m
    arrays, a sequence of n plot frames will be created. Each
    frame f will show a plot of the ydata[f,:] values against the
    xdata[f,:] values. Initially, only the first frame is shown.
    The user can start an animation sequence which will cycle
    through the successive frames in the plot window.

    >>> plot(harpproduct)

    where 'harpproduct' is an object of type harp.Product (see
    also the documentation for the HARP python interface), will
    display a specific 2D plot associated with the type of data
    contained in 'harpproduct'.

    >>> plot(harpproduct, value='variablename')

    will create a 2D plot for the specified variable of the HARP
    product.

    >>> w = plot(data1)
    >>> plot(data2, window=w)

    will plot data2 in the same window 'w' as used for data1,
    instead of opening a separate window.

    In addition to the 'window' keyword, the plot() function also
    accepts a number of other, comma-separated optional
    properties of the form 'property=value'. Their order is
    arbitrary, but they must always appear after any data
    arguments to the plot() function.

    Available plot() properties are:
    window, windowtitle, size, pos, title, xrange, yrange,
    xmin, xmax, ymin, ymax, xlog, ylog, xbase, ybase, xlabel,
    ylabel, xnumticks, ynumticks, numticks, showanimationtoolbar,
    showpropertypanel,
    value, name, lines, linewidth, stipplepattern, points,
    pointsize, color, opacity.

    """
    import wx
    from . import windowhandler as WindowHandler
    from visan.plot import PlotFrame

    # validate all arguments

    defaultProperties = dict()
    dataSetAttributes = dict()
    dataSetLocation = []
    xdata = None
    ydata = None
    value = None
    try:
        value = kwargs["value"]
    except KeyError:
        pass

    if len(args) > 0:
        if isinstance(args[0], harp.Product):
            if len(args) != 1:
                raise ValueError("plot() does not allow additional data arguments when the first argument "
                                 "is a HARP product")
            xdata, ydata, dataSetAttributes, dataSetLocation, defaultProperties = PlotDataForProduct(args[0], value)
        else:
            if value is not None:
                raise ValueError("parameter 'value' (%s) is only allowed when plotting a HARP product" % value)
            if len(args) > 2:
                raise ValueError("plot() takes at most two data arguments (%d given)" % len(args))
            elif len(args) > 1:
                xdata = args[0]
                if isinstance(xdata, harp.Variable):
                    xdata = xdata.data
                ydata = args[1]
                if isinstance(ydata, harp.Variable):
                    ydata = ydata.data
            else:
                ydata = args[0]
                if isinstance(ydata, harp.Variable):
                    ydata = ydata.data
    elif value is not None:
        raise ValueError("parameter 'value' (%s) is only allowed when plotting a HARP product" % value)

    knownproperties = ["value", "window", "windowtitle", "size", "pos", "title", "xrange", "yrange", "xmin", "xmax",
                       "ymin", "ymax", "xlog", "ylog", "xbase", "ybase", "xlabel", "ylabel", "xnumticks", "ynumticks",
                       "numticks", "showanimationtoolbar", "showpropertypanel", "name", "lines", "linewidth",
                       "stipplepattern", "points", "pointsize", "color", "opacity"]

    unknowns = [k for k in list(kwargs.keys()) if k not in knownproperties]

    if len(unknowns) == 1:
        raise TypeError("Invalid keyword: %s\n(Supported keywords are: %s)" %
                        (unknowns[0], ', '.join(knownproperties)))
    elif len(unknowns) > 1:
        raise TypeError("Invalid keywords: %s\n(Supported keywords are: %s)" %
                        (', '.join(unknowns), ', '.join(knownproperties)))

    # window
    window = kwargs.get("window")
    if window is not None:
        if not isinstance(window, PlotFrame):
            raise ValueError("parameter 'window' (%s) does not refer to a plot window" % window)
        if not window:
            raise ValueError("parameter 'window' refers to a window that no longer exists")

    # windowtitle
    windowtitle = kwargs.get("windowtitle")
    if windowtitle is not None:
        try:
            windowtitle = str(windowtitle)
        except TypeError:
            raise ValueError("parameter 'windowtitle' should be a string")

    # size
    size = kwargs.get("size")
    if size is not None:
        try:
            size = tuple(size)
        except TypeError:
            raise TypeError("parameter 'size' should be a 2-element sequence of numbers (was: '%s')" % str(size))
        try:
            x = float(size[0])
            y = float(size[1])
        except TypeError:
            raise TypeError("parameter 'size' should be a 2-element sequence of numbers (was: '%s')" % str(size))
        if x <= 0 or y <= 0:
            raise ValueError("parameter 'size' must contain positive numbers (was: '%s')" % str(size))
        xmax = wx.SystemSettings_GetMetric(wx.SYS_SCREEN_X)
        if x > xmax:
            raise ValueError("x component of 'size' parameter must not exceed maximum screen width ('%g' > '%g')" %
                             (x, xmax))
        ymax = wx.SystemSettings_GetMetric(wx.SYS_SCREEN_Y)
        if y > ymax:
            raise ValueError("y component of 'size' parameter must not exceed maximum screen heigth ('%g' > '%g')" %
                             (y, ymax))

    # pos
    pos = kwargs.get("pos")
    if pos is not None:
        try:
            pos = tuple(pos)
        except TypeError:
            raise TypeError("parameter 'pos' should be a 2-element sequence of numbers (was: '%s')" % str(pos))
        try:
            x = float(pos[0])
            y = float(pos[1])
        except TypeError:
            raise TypeError("parameter 'pos' should be a 2-element sequence of numbers (was: '%s')" % str(pos))
        if x < 0 or y < 0:
            raise ValueError("parameter 'pos' must contain positive numbers (was: '%s')" % str(pos))
        xmax = wx.SystemSettings_GetMetric(wx.SYS_SCREEN_X)
        if x > xmax:
            raise ValueError("x component of 'pos' parameter must not exceed maximum screen width ('%g' > '%g')" %
                             (x, xmax))
        ymax = wx.SystemSettings_GetMetric(wx.SYS_SCREEN_Y)
        if y > ymax:
            raise ValueError("y component of 'pos' parameter must not exceed maximum screen heigth ('%g' > '%g')" %
                             (y, ymax))

    # check mutual exclusive parameters
    if kwargs.get("xmin") is not None and kwargs.get("xrange") is not None:
        raise ValueError("it is not allowed to specify both 'xrange' and 'xmin' in a single call")
    if kwargs.get("xmax") is not None and kwargs.get("xrange") is not None:
        raise ValueError("it is not allowed to specify both 'xrange' and 'xmax' in a single call")
    if kwargs.get("ymin") is not None and kwargs.get("yrange") is not None:
        raise ValueError("it is not allowed to specify both 'yrange' and 'ymin' in a single call")
    if kwargs.get("ymax") is not None and kwargs.get("yrange") is not None:
        raise ValueError("it is not allowed to specify both 'yrange' and 'ymax' in a single call")
    if kwargs.get("numticks") is not None and kwargs.get("xnumticks") is not None:
        raise ValueError("it is not allowed to specify both 'numticks' and 'xnumticks' in a single call")
    if kwargs.get("numticks") is not None and kwargs.get("ynumticks") is not None:
        raise ValueError("it is not allowed to specify both 'numticks' and 'ynumticks' in a single call")

    # process arguments

    if window is None:
        if size is None:
            size = WindowHandler.GetDefaultSize()
        if pos is None:
            pos = WindowHandler.GetNextPosition(size)
        plot = PlotFrame(size=size, pos=pos)
    else:
        plot = window
        if size is not None:
            plot.SetSize(size)
        if pos is not None:
            plot.Move(pos)

    # prevent intermediate repaints of the plotwindow until we have set all properties
    plot.Freeze()

    try:
        # add data set (if data was given)
        dataSetId = None
        if len(args) > 0:
            dataSetId = plot.AddDataSet(xdata, ydata)
            for attr in dataSetAttributes:
                plot.AddDataSetAttribute(dataSetId, attr, dataSetAttributes[attr])
            if len(dataSetLocation) == 2:
                # TODO: make this work for animated plots
                plot.AddDataSetLocation(dataSetId, dataSetLocation[0], dataSetLocation[1])

        # set generic plot properties
        if windowtitle is not None:
            plot.SetTitle(windowtitle)
        if kwargs.get("title") is not None:
            plot.SetPlotTitle(kwargs.get("title"))
        elif "title" in defaultProperties:
            plot.SetPlotTitle(defaultProperties["title"])
        if kwargs.get("xrange") is not None:
            plot.SetXAxisRange(kwargs.get("xrange"))
        if kwargs.get("yrange") is not None:
            plot.SetYAxisRange(kwargs.get("yrange"))
        if kwargs.get("xmin") is not None:
            if kwargs.get("xmax") is not None:
                plot.SetXAxisRange([kwargs.get("xmin"), kwargs.get("xmax")])
            else:
                plot.SetXAxisRange([kwargs.get("xmin"), plot.GetXAxisRange()[1]])
        if kwargs.get("xmax") is not None:
            plot.SetXAxisRange([plot.GetXAxisRange()[0], kwargs.get("xmax")])
        if kwargs.get("ymin") is not None:
            if kwargs.get("ymax") is not None:
                plot.SetYAxisRange([kwargs.get("ymin"), kwargs.get("ymax")])
            else:
                plot.SetYAxisRange([kwargs.get("ymin"), plot.GetYAxisRange()[1]])
        if kwargs.get("ymax") is not None:
            plot.SetYAxisRange([plot.GetYAxisRange()[0], kwargs.get("ymax")])
        if kwargs.get("xlog") is not None:
            plot.SetXLogAxis(kwargs.get("xlog"))
        elif "xlog" in defaultProperties:
            plot.SetXLogAxis(defaultProperties["xlog"])
        if kwargs.get("ylog") is not None:
            plot.SetYLogAxis(kwargs.get("ylog"))
        elif "ylog" in defaultProperties:
            plot.SetYLogAxis(defaultProperties["ylog"])
        if kwargs.get("xbase") is not None:
            plot.SetXAxisBase(kwargs.get("xbase"))
        elif "xbase" in defaultProperties:
            plot.SetXAxisBase(defaultProperties["xbase"])
        if kwargs.get("ybase") is not None:
            plot.SetYAxisBase(kwargs.get("ybase"))
        elif "ybase" in defaultProperties:
            plot.SetYAxisBase(defaultProperties["ybase"])
        if kwargs.get("xlabel") is not None:
            plot.SetXAxisTitle(kwargs.get("xlabel"))
        elif "xlabel" in defaultProperties:
            plot.SetXAxisTitle(defaultProperties["xlabel"])
        if kwargs.get("ylabel") is not None:
            plot.SetYAxisTitle(kwargs.get("ylabel"))
        elif "ylabel" in defaultProperties:
            plot.SetYAxisTitle(defaultProperties["ylabel"])
        if kwargs.get("numticks") is not None:
            plot.SetXNumAxisLabels(kwargs.get("numticks"))
            plot.SetYNumAxisLabels(kwargs.get("numticks"))
        if kwargs.get("xnumticks") is not None:
            plot.SetXNumAxisLabels(kwargs.get("xnumticks"))
        if kwargs.get("ynumticks") is not None:
            plot.SetYNumAxisLabels(kwargs.get("ynumticks"))
        if kwargs.get("showanimationtoolbar") is not None:
            plot.ShowAnimationToolbar(kwargs.get("showanimationtoolbar"))
        elif "showanimationtoolbar" in defaultProperties:
            plot.ShowAnimationToolbar(defaultProperties["showanimationtoolbar"])
        if kwargs.get("showpropertypanel") is not None:
            plot.ShowPropertyPanel(kwargs.get("showpropertypanel"))
        elif "showpropertypanel" in defaultProperties:
            plot.ShowPropertyPanel(defaultProperties["showpropertypanel"])

        # set data set properties
        if dataSetId is not None:
            if kwargs.get("name") is not None:
                plot.SetDataSetLabel(dataSetId, kwargs.get("name"))
            elif "name" in defaultProperties:
                plot.SetDataSetLabel(dataSetId, defaultProperties["name"])
            if kwargs.get("lines") is not None:
                plot.SetPlotLines(dataSetId, kwargs.get("lines"))
            if kwargs.get("linewidth") is not None:
                plot.SetLineWidth(dataSetId, kwargs.get("linewidth"))
            if kwargs.get("stipplepattern") is not None:
                plot.SetLineStipplePattern(dataSetId, kwargs.get("stipplepattern"))
            if kwargs.get("points") is not None:
                plot.SetPlotPoints(dataSetId, kwargs.get("points"))
            if kwargs.get("pointsize") is not None:
                plot.SetPointSize(dataSetId, kwargs.get("pointsize"))
            if kwargs.get("color") is not None:
                plot.SetPlotColor(dataSetId, kwargs.get("color"))
            if kwargs.get("opacity") is not None:
                plot.SetOpacity(dataSetId, kwargs.get("opacity"))
            else:
                # use our own VISAN specific opacity setting
                plot.SetOpacity(dataSetId, 0.6)
    except Exception:
        if window is None:
            # only destroy the window if we created it ourselves
            plot.Destroy()
        raise
    plot.Thaw()

    return plot


def wplot(*args, **kwargs):
    """ Display a dataset plotted onto an Earth world map.

    The wplot() function accepts dataset parameters similar to
    the plot() function. It can plot basic longitude vs. latitude
    data as well as HARP products containing latitude/longitude
    information. All datasets are plotted onto a 2D or 3D
    Earth world map. For a full reference guide, please consult
    the VISAN documentation at:

        <http://www.stcorp.nl/beat/>

    Basic usage:

    >>> wplot(latitude, longitude)

    will plot the locations using the information from the 'latitude'
    and 'longitude' arrays and project these onto a world map, in a
    single plot frame. The type of projection can be chosen by
    supplying a value for the 'projection' keyword.
    The latitude/longitude arrays may contain either points or
    corner coordinates.

    >>> wplot(latitude, longitude, data)

    will plot the locations using 'data' as color information.

    If 'data' is one-dimensional, then the data will be plotted as
    points or swaths.
    It will be plotted as points if the latitude/longitude arrays
    are one-dimensional, and as swaths if latitude/longitude arrays
    are two-dimensional (in which case the second dimension should
    capture the corners of the swath)

    If 'data' has more than one dimension, then 'data' will be plotted
    as a grid using the latitude and longitude as axis.
    Latitude should match the last dimension of 'data' and longitude
    the second-last dimension.
    If 'data' is three dimensional, then an animation plot is created
    with the first dimension being used as time dimension.

    >>> plot(harpproduct)

    where 'harpproduct' is an object of type harp.Product (see
    also the documentation for the HARP python interface), will
    display a specific worldmap plot associated with the type of
    data contained in 'harpproduct'.

    >>> plot(harpproduct, value='variablename')

    will create a worldmap plot for the specified variable of the
    HARP product.

    The wplot() function also accepts a number of comma-separated
    optional properties of the form 'property=value'. Their order
    is arbitrary, but they must always appear after any data
    arguments to the wplot() function.

    Available wplot() properties are:
    window, windowtitle, size, pos, title, centerlat,
    centerlon, zoom, projection, projectionlat, projectionlon,
    showanimationtoolbar, showpropertypanel, showcolorbar,
    value, colortable, colorrange, colorbartitle, numcolorlabels,
    opacity, linewidth, pointsize, drawpath, drawlocation,
    heightfactor, minheightvalue, maxheightvalue, deltaradius.

    """
    import wx
    from . import windowhandler as WindowHandler
    from visan.plot import WorldPlotFrame

    kPointData = 0
    kSwathData = 1
    kGridData = 2

    # validate all arguments

    defaultProperties = dict()
    dataSetAttributes = dict()
    latitude = None
    longitude = None
    data = None
    value = None
    datatype = kPointData
    plotPoints = False
    plotLines = False
    try:
        value = kwargs["value"]
    except KeyError:
        pass

    if kwargs.get("drawpath") is not None:
        plotLines = bool(kwargs["drawpath"])

    if kwargs.get("drawlocation") is not None:
        plotPoints = bool(kwargs["drawlocation"])

    if len(args) > 0:
        if isinstance(args[0], harp.Product):
            if len(args) != 1:
                raise ValueError("wplot() does not allow additional data arguments when the first argument is a "
                                 "HARP product")
            datatype, data, latitude, longitude, dataSetAttributes, defaultProperties = \
                WorldPlotDataForProduct(args[0], plotPoints or plotLines, value)
        else:
            if value is not None:
                raise ValueError("parameter 'value' (%s) is only allowed when plotting a HARP product" % value)
            if len(args) == 3:
                data = args[2]
                if isinstance(data, harp.Variable):
                    data = data.data
            elif len(args) != 2:
                raise ValueError("wplot() takes either two or three data arguments if the first argument is not a "
                                 "HARP product (%d given)" % len(args))
            latitude = args[0]
            if isinstance(latitude, harp.Variable):
                latitude = latitude.data
            if not isinstance(latitude, numpy.ndarray):
                try:
                    latitude = numpy.asarray(args[0])
                except TypeError:
                    raise TypeError("latitude data argument cannot be converted to numpy array. %s" % str(latitude))
            longitude = args[1]
            if isinstance(longitude, harp.Variable):
                longitude = longitude.data
            if not isinstance(longitude, numpy.ndarray):
                try:
                    longitude = numpy.asarray(args[1])
                except TypeError:
                    raise TypeError("longitude data argument cannot be converted to numpy array. %s" % str(longitude))

            if data is not None:
                try:
                    data = numpy.asarray(data)
                except TypeError:
                    raise TypeError("data argument cannot be converted to numpy array. %s" % str(data))
                if data.ndim > 1:
                    datatype = kGridData

            if datatype == kPointData and latitude.ndim == 2 and longitude.ndim == 2:
                datatype = kSwathData

    elif value is not None:
        raise ValueError("parameter 'value' (%s) is only allowed when plotting a HARP product" % value)

    knownproperties = ["window", "windowtitle", "size", "pos", "title", "centerlat", "centerlon", "zoom", "projection",
                       "projectionlat", "projectionlon", "showanimationtoolbar", "showpropertypanel", "showcolorbar",
                       "value", "colortable", "colorrange", "colorbartitle", "numcolorlabels", "opacity", "linewidth",
                       "pointsize", "drawpath", "drawlocation", "heightfactor", "minheightvalue", "maxheightvalue",
                       "deltaradius"]

    unknowns = [k for k in list(kwargs.keys()) if k not in knownproperties]

    if len(unknowns) == 1:
        raise TypeError("Invalid keyword: %s\n(Supported keywords are: %s)" % (unknowns[0], ', '.join(knownproperties)))
    elif len(unknowns) > 1:
        raise TypeError("Invalid keywords: %s\n(Supported keywords are: %s)" %
                        (', '.join(unknowns), ', '.join(knownproperties)))

    # window
    window = kwargs.get("window")
    if window is not None:
        if not isinstance(window, WorldPlotFrame):
            raise ValueError("parameter 'window' (%s) does not refer to a wplot window" % window)
        if not window:
            raise ValueError("parameter 'window' refers to a window that no longer exists")

    # windowtitle
    windowtitle = kwargs.get("windowtitle")
    if windowtitle is not None:
        try:
            windowtitle = str(windowtitle)
        except TypeError:
            raise ValueError("parameter 'windowtitle' should be a string")

    # size
    size = kwargs.get("size")
    if size is not None:
        try:
            size = tuple(size)
        except TypeError:
            raise TypeError("parameter 'size' should be a 2-element sequence of numbers (was: '%s')" % str(size))
        try:
            x = float(size[0])
            y = float(size[1])
        except TypeError:
            raise TypeError("parameter 'size' should be a 2-element sequence of numbers (was: '%s')" % str(size))
        if x <= 0 or y <= 0:
            raise ValueError("parameter 'size' must contain positive numbers (was: '%s')" % str(size))
        xmax = wx.SystemSettings_GetMetric(wx.SYS_SCREEN_X)
        if x > xmax:
            raise ValueError("x component of 'size' parameter must not exceed maximum screen width ('%g' > '%g')" %
                             (x, xmax))
        ymax = wx.SystemSettings_GetMetric(wx.SYS_SCREEN_Y)
        if y > ymax:
            raise ValueError("y component of 'size' parameter must not exceed maximum screen heigth ('%g' > '%g')" %
                             (y, ymax))

    # pos
    pos = kwargs.get("pos")
    if pos is not None:
        try:
            pos = tuple(pos)
        except TypeError:
            raise TypeError("parameter 'pos' should be a 2-element sequence of numbers (was: '%s')" % str(pos))
        try:
            x = float(pos[0])
            y = float(pos[1])
        except TypeError:
            raise TypeError("parameter 'pos' should be a 2-element sequence of numbers (was: '%s')" % str(pos))
        if x < 0 or y < 0:
            raise ValueError("parameter 'pos' must contain positive numbers (was: '%s')" % str(pos))
        xmax = wx.SystemSettings_GetMetric(wx.SYS_SCREEN_X)
        if x > xmax:
            raise ValueError("x component of 'pos' parameter must not exceed maximum screen width ('%g' > '%g')" %
                             (x, xmax))
        ymax = wx.SystemSettings_GetMetric(wx.SYS_SCREEN_Y)
        if y > ymax:
            raise ValueError("y component of 'pos' parameter must not exceed maximum screen heigth ('%g' > '%g')" %
                             (y, ymax))

    # process arguments

    if window is None:
        if size is None:
            size = WindowHandler.GetDefaultSize()
        if pos is None:
            pos = WindowHandler.GetNextPosition(size)
        plot = WorldPlotFrame(size=size, pos=pos)
    else:
        plot = window
        if size is not None:
            plot.SetSize(size)
        if pos is not None:
            plot.Move(pos)

    # prevent intermediate repaints of the plotwindow until we have set all properties
    plot.Freeze()

    try:
        # add data set (if data was given)
        dataSetId = None
        if len(args) > 0:
            if datatype == kPointData:
                if plotLines:
                    dataSetId = plot.AddLineData(latitude, longitude)
                else:
                    dataSetId = plot.AddPointData(latitude, longitude, data)
            elif datatype == kSwathData:
                dataSetId = plot.AddSwathData(latitude, longitude, data)
            elif datatype == kGridData:
                dataSetId = plot.AddGridData(latitude, longitude, data)
            else:
                raise AssertionError("invalid datatype (%d) for plotdata" % datatype)
            for attr in dataSetAttributes:
                plot.AddDataSetAttribute(dataSetId, attr, dataSetAttributes[attr])

        # set generic plot properties
        if windowtitle is not None:
            plot.SetTitle(windowtitle)
        if kwargs.get("title") is not None:
            plot.SetPlotTitle(kwargs.get("title"))
        elif "title" in defaultProperties:
            plot.SetPlotTitle(defaultProperties["title"])
        if kwargs.get("showanimationtoolbar") is not None:
            plot.ShowAnimationToolbar(kwargs.get("showanimationtoolbar"))
        elif "showanimationtoolbar" in defaultProperties:
            plot.ShowAnimationToolbar(defaultProperties["showanimationtoolbar"])
        if kwargs.get("showpropertypanel") is not None:
            plot.ShowPropertyPanel(kwargs.get("showpropertypanel"))
        elif "showpropertypanel" in defaultProperties:
            plot.ShowPropertyPanel(defaultProperties["showpropertypanel"])
        if kwargs.get("showcolorbar") is not None:
            plot.ShowColorBar(kwargs.get("showcolorbar"))
        elif "showcolorbar" in defaultProperties:
            plot.ShowColorBar(defaultProperties["showcolorbar"])
        if kwargs.get("projection") is not None:
            plot.SetProjection(kwargs.get("projection"))
        if kwargs.get("projectionlat") is not None:
            plot.SetProjectionCenterLatitude(kwargs.get("projectionlat"))
        if kwargs.get("projectionlon") is not None:
            plot.SetProjectionCenterLongitude(kwargs.get("projectionlon"))
        if kwargs.get("centerlat") is not None:
            plot.SetViewCenterLatitude(kwargs.get("centerlat"))
        if kwargs.get("centerlon") is not None:
            plot.SetViewCenterLongitude(kwargs.get("centerlon"))
        if kwargs.get("zoom") is not None:
            plot.SetViewZoom(kwargs.get("zoom"))
        else:
            if window is None:
                # if we are creating the plot, set the initial zoom level (if 1.0 is not appropriate)
                if kwargs.get("projection") is None or kwargs.get("projection").lower() == "3d":
                    plot.SetViewZoom(2.5)
                elif kwargs.get("projection").lower() == "lambert cylindrical" or \
                        kwargs.get("projection").lower() == "plate caree" or \
                        kwargs.get("projection").lower() == "mollweide" or \
                        kwargs.get("projection").lower() == "robinson":
                    plot.SetViewZoom(1.6)

        # set data set properties
        if dataSetId is not None:
            if kwargs.get("name") is not None:
                plot.SetDataSetLabel(dataSetId, kwargs.get("name"))
            elif "name" in defaultProperties:
                plot.SetDataSetLabel(dataSetId, defaultProperties["name"])
            if kwargs.get("colortable") is not None:
                try:
                    plot.GetColorTable(dataSetId).SetColorTableByName(kwargs.get("colortable"))
                except Exception:
                    plot.GetColorTable(dataSetId).Import(kwargs.get("colortable"))
            elif "colortable" in defaultProperties:
                plot.GetColorTable(dataSetId).SetColorTableByName(defaultProperties["colortable"])
            if kwargs.get("colorrange") is not None:
                plot.SetColorRange(dataSetId, kwargs.get("colorrange"))
            elif "colorrange" in defaultProperties:
                plot.SetColorRange(dataSetId, defaultProperties["colorrange"])
            if kwargs.get("colorbartitle") is not None:
                plot.SetColorBarTitle(dataSetId, kwargs.get("colorbartitle"))
            elif "colorbartitle" in defaultProperties:
                plot.SetColorBarTitle(dataSetId, defaultProperties["colorbartitle"])
            if kwargs.get("numcolorlabels") is not None:
                plot.SetNumColorBarLabels(dataSetId, kwargs.get("numcolorlabels"))
            if kwargs.get("opacity") is not None:
                plot.SetOpacity(dataSetId, kwargs.get("opacity"))
            elif "opacity" in defaultProperties:
                plot.SetOpacity(dataSetId, defaultProperties["opacity"])
            if kwargs.get("linewidth") is not None:
                plot.SetLineWidth(dataSetId, kwargs.get("linewidth"))
            if kwargs.get("pointsize") is not None:
                plot.SetPointSize(dataSetId, kwargs.get("pointsize"))
            if kwargs.get("heightfactor") is not None:
                plot.SetHeightFactor(dataSetId, kwargs.get("heightfactor"))
            if kwargs.get("minheightvalue") is not None:
                plot.SetMinHeightValue(dataSetId, kwargs.get("minheightvalue"))
            if kwargs.get("maxheightvalue") is not None:
                plot.SetMinHeightValue(dataSetId, kwargs.get("maxheightvalue"))
            if kwargs.get("deltaradius") is not None:
                try:
                    deltaradius = float(kwargs.get("deltaradius"))
                except TypeError:
                    raise TypeError("deltaradius property should be a float (was: '%s')" %
                                    str(kwargs.get("deltaradius")))
                plot.SetReferenceHeight(dataSetId, 1.0 + deltaradius)
    except Exception:
        if window is None:
            # only destroy the window if we created it ourselves
            plot.Destroy()
        raise
    plot.Thaw()

    # TODO: Is this still applicable?
    # The Yield and Refresh are needed to make sure that colortable changes are shown.
    # Somehow the plot is not rendered using the new colortable after the Thaw.
    # We have to let the current Refresh result in a Render (using wx.Yield())
    # and then ask for a new render using Refresh()
    wx.Yield()
    plot.Refresh()

    return plot


def executescript(filename, globals=None, mainargs=None):
    """ Execute a script.

    This routine runs an external python script.

    >>> executescript(filename)

    runs the script file located at the path indicated by 'filename' and returns
    to the prompt when the script is finished. If the system is unable to locate
    the file indicated by 'filename' an execption will be thrown.

    When you run a script, VISAN will automatically add the location of your
    script to the module searchpath (sys.path). This means that if your script
    uses modules that are located in the same directory as the script file then
    these modules will automatically be found. When the script is finished VISAN
    will change the module searchpath back to its original value.


    If you want the script to have access to variables that you have already
    set or if you want to have access to variables that are going to be set
    by the script, just pass 'globals()' as second parameter:

    >>> executescript(filename, globals())


    If the script needs to be run as a 'main' script (i.e. as if it was run as an
    argument to the python executable) and/or requires command line arguments you
    can use the 'mainargs' keyword. The 'mainargs' keyword should contain a list
    of command line options (which may be an empty list). For instance, to run a
    setup.py script to install an external module you could use:

    >>> executescript('/home/user/package/setup.py', mainargs = ['install'])

    This will have the result the script is called with the '__name__
    variable set to '__main__' and sys.argv will have been set to
    ['/home/user/package/setup.py', 'install'].
    The current working directory for the script will also be set to the
    directory of the script (e.g. '/home/user/package').

    Note that if a 'mainargs' keyword parameter is provided, the 'globals'
    parameter to executescript() (if present) will be ignored.

    """
    if os.path.exists(filename):
        if os.path.isfile(filename):
            print(("Executing VISAN/Python script '%s':\n" % filename))

            oldcwd = os.getcwd()
            scriptdir = os.path.dirname(filename)
            sys.path.insert(0, scriptdir)
            oldargv = sys.argv
            try:
                if mainargs is None:
                    if globals is None:
                        exec(compile(open(filename).read(), filename, 'exec'))
                    else:
                        exec(compile(open(filename).read(), filename, 'exec'), globals)
                else:
                    sys.argv = [filename] + mainargs
                    os.chdir(scriptdir)
                    exec(compile(open(filename).read(), filename, 'exec'), {'__name__': '__main__'})
            finally:
                del sys.path[0]
                sys.argv = oldargv
                os.chdir(oldcwd)
        else:
            raise IOError("Error executing VISAN/Python script, the path '%s' does not point to a file\n" % filename)
    else:
        raise IOError("Error executing VISAN/Python script, the file '%s' does not exist\n" % filename)


def browseproduct(filename):
    """ Open the Product Browser with a specific product.

    This routine opens a new Product Browser window with the give product file.

    >>> browseproduct(filename)

    If the system is unable to locate the file indicated by 'filename' an
    execption will be thrown.

    """
    if os.path.exists(filename):
        if os.path.isfile(filename):
            import wx
            wx.GetApp().BrowseProduct(filename)
        else:
            raise IOError("Error opening product file, the path '%s' does not point to a file\n" % filename)
    else:
        raise IOError("Error opening product file, the file '%s' does not exist\n" % filename)


def version():
    """ Get version number of VISAN.

    >>> v = version()

    returns the version number of the current VISAN release.
    """
    from . import VERSION
    return VERSION



# This is an example VISAN script for the SCI_NL__1P product

# Make sure to set the 'products-file directory' option in the VISAN Preferences panel to
# a directory containing SCI_NL__1P products.

# This example will then take the first product it finds in this directory and
# for that product plot the measured limb spectra for the range 290nm - 450nm


def run():

    import glob
    import wx

    productdir = str(wx.Config.Get().Read('DirectoryLocation/Products'))

    # Use glob to find all files in productdir starting with 'SCI_NL__1P' and ending with '.child'
    files = glob.glob(os.path.join(productdir, "SCI_NL__1P*.child"))

    if len(files) == 0:
        print(("Could not find any SCI_NLC_1P files in directory '" + productdir + "'"))
        return

    # We only ingest the first file from the list
    record = harp.import_product(files[0], "wavelength>=290;wavelength<=450", "data=limb")

    plot(record, showpropertypanel=True, name=os.path.basename(files[0]),
         title="SCI_NLC_1P spectral readout example (limb: 290-450nm)")
    wplot(record, colortable='RedToGreen', projection="Mollweide")


run()



# This is an example VISAN script for the GOM_LIM_1P product

# Make sure to set the 'products-file directory' option in the VISAN Preferences panel to
# a directory containing GOM_LIM_1P products.

# This example will then take the first product it finds in this directory and
# for that product plot the background spectra


def run():

    import glob
    import wx

    productdir = str(wx.Config.Get().Read('DirectoryLocation/Products'))

    # Use glob to find all files in productdir starting with 'GOM_LIM_1P'
    files = glob.glob(os.path.join(productdir, "GOM_LIM_1P*"))
    if len(files) == 0:
        print(("Could not find any GOM_LIM_1P files in directory '" + productdir + "'"))
        return

    # We only ingest the first file from the list
    corup = harp.import_product(files[0], "", "spectra=upper")
    nocorup = harp.import_product(files[0], "", "spectra=upper;corrected=false")
    corlow = harp.import_product(files[0], "", "spectra=lower")
    nocorlow = harp.import_product(files[0], "", "spectra=lower;corrected=false")

    w = plot(nocorup, name="upper without correction")
    plot(corup, name="upper with correction", window=w)
    plot(nocorlow, name="lower without correction", window=w)
    plot(corlow, name="lower with correction", window=w, title="GOM_LIM_1P background spectra example")


run()



# Copyright (C) 2002-2022 S[&]T, The Netherlands.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import wx

from .labeledtextctrl import LabeledTextCtrl, EVT_VALUE_CHANGED
from .util import DetermineCharSize
from .validators import FloatValidator
from .worldplotwindow import AZIMUTHAL_PROJECTIONS, CYLINDRICAL_PROJECTIONS, PROJECTIONS


class WorldPlotPropertyPanel(wx.Panel):

    def __init__(self, parent, worldplotWindow):
        panelstyle = wx.TAB_TRAVERSAL
        if wx.Platform == '__WXGTK__':
            panelstyle |= wx.SUNKEN_BORDER
        wx.Panel.__init__(self, parent, -1, style=panelstyle)

        self.plotWindow = worldplotWindow

        # Create and configure all widgets
        self.CreateControls()
        self.CreateLayout()

    def CreateControls(self):
        (charWidth, charHeight) = DetermineCharSize(self)

        self.titleCtrl = LabeledTextCtrl(self, -1, label="Plot Title:", value=self.plotWindow.GetPlotTitle(),
                                         size=(10 * charWidth, -1), style=wx.TE_PROCESS_ENTER)
        self.titleCtrl.SetToolTip(wx.ToolTip("The plot title will be displayed centered at the top of the plot."))
        self.titleCtrl.Bind(EVT_VALUE_CHANGED, self.OnTitle)

        # static boxes should be created before the controls they contain in order to preserve the correct Z-Order
        self.projCenterBox = wx.StaticBox(self, -1, "Center of Projection")
        self.viewCtrlBox = wx.StaticBox(self, -1, "View Center and Zoom")

        self.projectionLabel = wx.StaticText(self, -1, "Projection:", style=wx.ALIGN_RIGHT)

        self.projectionChoice = wx.Choice(self, -1, choices=PROJECTIONS)
        self.projectionChoice.SetToolTip(wx.ToolTip("Projection to use for mapping the world globe to the "
                                                    "2-dimensional plane."))
        self.projectionChoice.Bind(wx.EVT_CHOICE, self.OnProjectionChoice)

        self.projCenterLatitudeCtrl = LabeledTextCtrl(self, -1, label="Lat:", validator=FloatValidator(),
                                                      formatstring="%-#.4g", size=(5 * charWidth, -1),
                                                      style=wx.TE_PROCESS_ENTER)
        self.projCenterLatitudeCtrl.SetToolTip(wx.ToolTip("Latitude of the projection's center point."))
        self.projCenterLatitudeCtrl.Bind(EVT_VALUE_CHANGED, self.OnProjCenterLatitude)

        self.projCenterLongitudeCtrl = LabeledTextCtrl(self, -1, label="Lon:", validator=FloatValidator(),
                                                       formatstring="%-#.4g", size=(5 * charWidth, -1),
                                                       style=wx.TE_PROCESS_ENTER)
        self.projCenterLongitudeCtrl.SetToolTip(wx.ToolTip("Longitude of the projection's center point."))
        self.projCenterLongitudeCtrl.Bind(EVT_VALUE_CHANGED, self.OnProjCenterLongitude)

        self.viewCenterLatitudeCtrl = LabeledTextCtrl(self, -1, label="Lat:", validator=FloatValidator(),
                                                      formatstring="%#.4g", size=(5 * charWidth, -1),
                                                      style=wx.TE_PROCESS_ENTER)
        self.viewCenterLatitudeCtrl.SetToolTip(wx.ToolTip("Latitude of the view center point."))
        self.viewCenterLatitudeCtrl.Bind(EVT_VALUE_CHANGED, self.OnViewCenterLatitude)

        self.viewCenterLongitudeCtrl = LabeledTextCtrl(self, -1, label="Lon:", validator=FloatValidator(),
                                                       formatstring="%#.4g", size=(5 * charWidth, -1),
                                                       style=wx.TE_PROCESS_ENTER)
        self.viewCenterLongitudeCtrl.SetToolTip(wx.ToolTip("Latitude of the view center point."))
        self.viewCenterLongitudeCtrl.Bind(EVT_VALUE_CHANGED, self.OnViewCenterLongitude)

        self.zoomScaleCtrl = LabeledTextCtrl(self, -1, label="Zoom:", validator=FloatValidator(), formatstring="%#.3g",
                                             size=(5 * charWidth, -1), style=wx.TE_PROCESS_ENTER)
        self.zoomScaleCtrl.SetToolTip(wx.ToolTip("View scale around the view center."))
        self.zoomScaleCtrl.Bind(EVT_VALUE_CHANGED, self.OnZoomScale)

        self.UpdateControls()

    def CreateLayout(self):
        spacing = 10
        if wx.Platform == '__WXMAC__':
            spacing = 6

        self.hsizer = wx.BoxSizer(wx.HORIZONTAL)
        self.hsizer.Add(self.projectionLabel,  0, wx.ALIGN_CENTER | wx.RIGHT, border=spacing)
        self.hsizer.Add(self.projectionChoice, 0, wx.ALIGN_CENTER)

        self.hsizer2 = wx.BoxSizer(wx.HORIZONTAL)
        self.hsizer2.Add((spacing, 0), 0)
        self.hsizer2.Add(self.projCenterLatitudeCtrl, 0, wx.RIGHT, border=spacing)
        self.hsizer2.Add(self.projCenterLongitudeCtrl, 0)

        self.hsizer3 = wx.BoxSizer(wx.HORIZONTAL)
        self.hsizer3.Add((spacing, 0), 0)
        self.hsizer3.Add(self.viewCenterLatitudeCtrl, 0, wx.RIGHT, border=spacing)
        self.hsizer3.Add(self.viewCenterLongitudeCtrl, 0)

        self.hsizer4 = wx.BoxSizer(wx.HORIZONTAL)
        self.hsizer4.Add((spacing, 0), 0)
        self.hsizer4.Add(self.zoomScaleCtrl, 0, wx.RIGHT, border=spacing)

        self.vsizer = wx.StaticBoxSizer(self.projCenterBox, wx.VERTICAL)
        self.vsizer.Add((0, spacing), 0)
        self.vsizer.Add(self.hsizer2, 1, wx.LEFT | wx.RIGHT, border=spacing // 2)
        self.vsizer.Add((0, spacing), 0)

        self.vsizer2 = wx.StaticBoxSizer(self.viewCtrlBox, wx.VERTICAL)
        self.vsizer2.Add(self.hsizer3, 1, wx.LEFT | wx.RIGHT, border=spacing // 2)
        self.vsizer2.Add((0, spacing), 0)
        self.vsizer2.Add(self.hsizer4, 1, wx.LEFT | wx.RIGHT, border=spacing // 2)
        self.vsizer2.Add((0, spacing), 0)

        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.sizer.Add(self.titleCtrl, 0, wx.EXPAND | wx.ALL, border=10)
        self.sizer.Add((0, spacing), 0)
        self.sizer.Add(self.hsizer, 0, wx.EXPAND)
        self.sizer.Add((0, spacing), 0)
        self.sizer.Add(self.vsizer, 0, wx.EXPAND)
        self.sizer.Add((0, spacing), 0)
        self.sizer.Add(self.vsizer2,  0, wx.EXPAND)

        self.mainsizer = wx.BoxSizer(wx.VERTICAL)
        self.mainsizer.Add(self.sizer, 1, wx.EXPAND | wx.ALL, border=spacing)

        self.SetSizerAndFit(self.mainsizer)

    def UpdateControls(self):
        # Retrieve and set the values of axis properties
        self.titleCtrl.SetValue(self.plotWindow.GetPlotTitle())
        projection = PROJECTIONS.index(self.plotWindow.GetProjection())
        self.projectionChoice.SetSelection(projection)
        self.projCenterLatitude = self.plotWindow.GetProjectionCenterLatitude()
        self.projCenterLatitudeCtrl.SetValue(self.projCenterLatitude)
        self.projCenterLongitude = self.plotWindow.GetProjectionCenterLongitude()
        self.projCenterLongitudeCtrl.SetValue(self.projCenterLongitude)

        self.viewCenterLatitude = self.plotWindow.GetViewCenterLatitude()
        self.viewCenterLatitudeCtrl.SetValue(self.viewCenterLatitude)
        self.viewCenterLongitude = self.plotWindow.GetViewCenterLongitude()
        self.viewCenterLongitudeCtrl.SetValue(self.viewCenterLongitude)
        self.zoomScale = self.plotWindow.GetViewZoom()
        self.zoomScaleCtrl.SetValue(self.zoomScale)

        # Update enabled state of text fields
        self.projCenterLatitudeCtrl.Enable(PROJECTIONS[projection] in AZIMUTHAL_PROJECTIONS)
        self.projCenterLongitudeCtrl.Enable(PROJECTIONS[projection] in CYLINDRICAL_PROJECTIONS or
                                            PROJECTIONS[projection] in AZIMUTHAL_PROJECTIONS)

    def OnTitle(self, event):
        self.plotWindow.SetPlotTitle(event.value)

    def OnProjectionChoice(self, event):
        self.plotWindow.SetProjection(PROJECTIONS[self.projectionChoice.GetSelection()])
        self.UpdateControls()

    def OnProjCenterLatitude(self, event):
        try:
            value = float(self.projCenterLatitudeCtrl.GetValue())
        except ValueError:
            wx.Bell()
            self.projCenterLatitudeCtrl.SetValue(self.projCenterLatitude)
            return
        if value < -90 or value > 90:
            wx.Bell()
            self.projCenterLatitudeCtrl.SetValue(self.projCenterLatitude)
            return
        self.projCenterLatitude = value
        self.plotWindow.SetProjectionCenterLatitude(self.projCenterLatitude)

    def OnProjCenterLongitude(self, event):
        try:
            value = float(self.projCenterLongitudeCtrl.GetValue())
        except ValueError:
            wx.Bell()
            self.projCenterLongitudeCtrl.SetValue(self.projCenterLongitude)
            return
        if value < -180 or value > 180:
            wx.Bell()
            self.projCenterLongitudeCtrl.SetValue(self.projCenterLongitude)
            return
        self.projCenterLongitude = value
        self.plotWindow.SetProjectionCenterLongitude(self.projCenterLongitude)

    def OnZoomScale(self, event):
        try:
            value = float(self.zoomScaleCtrl.GetValue())
        except ValueError:
            wx.Bell()
            self.zoomScaleCtrl.SetValue(self.zoomScale)
            return
        if value < 1.0:
            wx.Bell()
            self.zoomScaleCtrl.SetValue(self.zoomScale)
            return
        self.zoomScale = value
        self.plotWindow.SetViewZoom(self.zoomScale)

    def OnViewCenterLatitude(self, event):
        try:
            value = float(self.viewCenterLatitudeCtrl.GetValue())
        except ValueError:
            wx.Bell()
            self.viewCenterLatitudeCtrl.SetValue(self.viewCenterLatitude)
            return
        if value < -90 or value > 90:
            wx.Bell()
            self.viewCenterLatitudeCtrl.SetValue(self.viewCenterLatitude)
            return

        self.viewCenterLatitude = value
        self.plotWindow.SetViewCenter(self.viewCenterLatitude, self.viewCenterLongitude)

    def OnViewCenterLongitude(self, event):
        try:
            value = float(self.viewCenterLongitudeCtrl.GetValue())
        except ValueError:
            wx.Bell()
            self.viewCenterLongitudeCtrl.SetValue(self.viewCenterLongitude)
            return
        if value < -180 or value > 180:
            wx.Bell()
            self.viewCenterLongitudeCtrl.SetValue(self.viewCenterLongitude)
            return

        self.viewCenterLongitude = value
        self.plotWindow.SetViewCenter(self.viewCenterLatitude, self.viewCenterLongitude)



# This is an example VISAN script for the OMI_L2 O3 TOMS and DOAS products

# Make sure to set the 'products-file directory' option in the VISAN Preferences panel to
# a directory containing OMI_L2 O3 TOMS or DOAS products.

# This example will then take the first product it finds in this directory and
# for that product plot the total ozone columns


def run():

    import glob
    import wx

    productdir = str(wx.Config.Get().Read('DirectoryLocation/Products'))

    # Use glob to find all OMI O3 files in productdir. We first try to find O3
    # TOMS files and if that fails we try to find the O3 DOAS files.
    files = glob.glob(os.path.join(productdir, "OMI-Aura_L2-OMTO3*"))
    if len(files) == 0:
        files = glob.glob(os.path.join(productdir, "OMI-Aura_L2-OMDOAO3*"))
    if len(files) == 0:
        print(("Could not find any OMTO3/OMDOAO3 files in directory '" + productdir + "'"))
        return

    # Ingest the total ozone column values from the first OMI L2 O3 file from
    # the list.
    o3 = harp.import_product(files[0])
    w = wplot(o3, colortable="Ozone", colorrange=(150.0, 500.0), numcolorlabels=8)


run()



# Copyright (C) 2002-2022 S[&]T, The Netherlands.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import wx

from .labeledtextctrl import LabeledTextCtrl, EVT_VALUE_CHANGED
from .util import DetermineCharSize
from .validators import DigitValidator, FloatValidator


class AxisPropertyPanel(wx.Panel):

    def __init__(self, parent, plotWindow, axisId):
        panelstyle = wx.TAB_TRAVERSAL
        if wx.Platform == '__WXGTK__':
            panelstyle |= wx.SUNKEN_BORDER
        wx.Panel.__init__(self, parent, -1, style=panelstyle)

        self.plotWindow = plotWindow
        self.maxTicks = 50
        self.axisId = axisId

        # Create and configure all widgets
        self.CreateControls()
        self.CreateLayout()

    def CreateControls(self):
        (charWidth, charHeight) = DetermineCharSize(self)

        # static boxes should be created before the controls they contain in order to preserve the correct Z-Order
        self.rangeBox = wx.StaticBox(self, -1, "Range")

        self.TitleCtrl = LabeledTextCtrl(self, -1, label="Label:", size=(8 * charWidth, -1), style=wx.TE_PROCESS_ENTER)
        self.TitleCtrl.SetValue(self.plotWindow.GetAxisTitle(self.axisId))
        self.TitleCtrl.SetToolTip(wx.ToolTip("The axis title will be displayed centered on the outside of the axis "
                                             "in the plot panel."))
        self.TitleCtrl.Bind(EVT_VALUE_CHANGED, self.OnTitle)

        self.LogAxisCtrl = wx.CheckBox(self, -1, "Logarithmic Scale")
        self.LogAxisCtrl.SetValue(self.plotWindow.GetLogAxis(self.axisId))
        self.LogAxisCtrl.SetToolTip(wx.ToolTip("Use a logarithmic axis. Disabled if the current range of axis values "
                                               "contains the number 0."))
        self.LogAxisCtrl.Bind(wx.EVT_CHECKBOX, self.OnLog)

        self.NTicksCtrl = LabeledTextCtrl(self, -1, label="Nr of Ticks:", validator=DigitValidator(),
                                          size=(3 * charWidth, -1), style=wx.TE_PROCESS_ENTER)
        self.NTicksCtrl.SetValue(self.plotWindow.GetNumAxisLabels(self.axisId))
        self.NTicksCtrl.SetToolTip(wx.ToolTip("The target number of ticks to display on the axis. "
                                              "The actual number of ticks is calculated; this control's value is "
                                              "what the calculation will try to aim for."))
        self.NTicksCtrl.Bind(EVT_VALUE_CHANGED, self.OnNTicks)

        self.BaseCtrl = LabeledTextCtrl(self, -1, label="Base:", validator=FloatValidator(), formatstring="%-#.3g",
                                        size=(5 * charWidth, -1), style=wx.TE_PROCESS_ENTER)
        self.BaseCtrl.SetValue(self.plotWindow.GetAxisBase(self.axisId))
        self.BaseCtrl.SetToolTip(wx.ToolTip("The base for the linear and logarithmic tick calculation for the axis. "
                                            "This value can be fractional but must be greater than 1."))
        self.BaseCtrl.Bind(EVT_VALUE_CHANGED, self.OnBase)

        self.MinCtrl = LabeledTextCtrl(self, -1, label="Min:", validator=FloatValidator(), formatstring="%-#.3g",
                                       size=(5 * charWidth, -1), style=wx.TE_PROCESS_ENTER)
        self.MinCtrl.SetValue(self.plotWindow.GetAxisRangeMin(self.axisId))
        self.MinCtrl.SetToolTip(wx.ToolTip("The currently displayed minimum range value."))
        self.MinCtrl.Bind(EVT_VALUE_CHANGED, self.OnMin)

        self.MaxCtrl = LabeledTextCtrl(self, -1, label="Max:", validator=FloatValidator(), formatstring="%-#.3g",
                                       size=(5 * charWidth, -1), style=wx.TE_PROCESS_ENTER)
        self.MaxCtrl.SetValue(self.plotWindow.GetAxisRangeMax(self.axisId))
        self.MaxCtrl.SetToolTip(wx.ToolTip("The currently displayed maximum range value."))
        self.MaxCtrl.Bind(EVT_VALUE_CHANGED, self.OnMax)

    def CreateLayout(self):

        sizer0 = wx.BoxSizer(wx.HORIZONTAL)
        sizer0.Add(self.NTicksCtrl, 0, wx.RIGHT, border=10)
        sizer0.Add(self.BaseCtrl, 0)

        rsizer1 = wx.BoxSizer(wx.HORIZONTAL)
        rsizer1.Add((15, 0), 0)
        rsizer1.Add(self.MinCtrl, 0, wx.RIGHT, border=10)
        rsizer1.Add(self.MaxCtrl, 0)

        sizer1 = wx.StaticBoxSizer(self.rangeBox, wx.VERTICAL)
        sizer1.Add((0, 10), 0)
        sizer1.Add(rsizer1, 1, wx.LEFT | wx.RIGHT, border=5)
        sizer1.Add((0, 10), 0)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.TitleCtrl, 0, wx.EXPAND)
        sizer.Add((0, 10), 0)
        sizer.Add(self.LogAxisCtrl, 0)
        sizer.Add((0, 10), 0)
        sizer.Add(sizer0, 0, wx.EXPAND)
        sizer.Add((0, 15), 0)
        sizer.Add(sizer1, 0, wx.EXPAND)

        mainSizer = wx.BoxSizer(wx.VERTICAL)
        mainSizer.Add(sizer, 1, wx.EXPAND | wx.ALL, border=10)

        self.SetSizerAndFit(mainSizer)

    def UpdateControls(self):
        # Retrieve and set the values of axis properties
        self.TitleCtrl.SetValue(self.plotWindow.GetAxisTitle(self.axisId))
        self.NTicksCtrl.SetValue(self.plotWindow.GetNumAxisLabels(self.axisId))
        self.LogAxisCtrl.SetValue(self.plotWindow.GetLogAxis(self.axisId))
        self.BaseCtrl.SetValue(self.plotWindow.GetAxisBase(self.axisId))
        self.UpdateRangeFromPlot()

    def UpdateRangeFromPlot(self):
        self.MinCtrl.SetValue(self.plotWindow.GetAxisRangeMin(self.axisId))
        self.MaxCtrl.SetValue(self.plotWindow.GetAxisRangeMax(self.axisId))

    def OnLog(self, event):
        curLog = self.plotWindow.GetLogAxis(self.axisId)
        newLog = self.LogAxisCtrl.IsChecked()

        # Check if the new value is different
        if newLog != curLog:
            self.plotWindow.SetLogAxis(self.axisId, newLog)
            self.UpdateRangeFromPlot()

    def OnBase(self, event):
        curBase = self.plotWindow.GetAxisBase(self.axisId)
        try:
            newBase = float(self.BaseCtrl.GetValue())
        except ValueError:
            wx.Bell()
            self.BaseCtrl.SetValue(curBase)
            return

        # Check if the new value is different
        if newBase != curBase:
            if newBase > 1:
                self.plotWindow.SetAxisBase(self.axisId, newBase)
            else:
                wx.Bell()
                self.BaseCtrl.SetValue(curBase)

    def OnTitle(self, event):
        curTitle = str(self.plotWindow.GetAxisTitle(self.axisId))
        newTitle = self.TitleCtrl.GetValue()

        # Check if the new value is different
        if newTitle != curTitle:
            self.plotWindow.SetAxisTitle(self.axisId, newTitle)

    def OnNTicks(self, event):
        curTicks = self.plotWindow.GetNumAxisLabels(self.axisId)
        try:
            newTicks = int(self.NTicksCtrl.GetValue())
        except ValueError:
            wx.Bell()
            self.NTicksCtrl.SetValue(curTicks)
            return

        # Check if the new value is different
        if newTicks != curTicks:
            if 0 < newTicks <= self.maxTicks:
                self.plotWindow.SetNumAxisLabels(self.axisId, newTicks)
            else:
                wx.Bell()
                if newTicks > self.maxTicks:
                    self.NTicksCtrl.SetValue(self.maxTicks)
                else:
                    self.NTicksCtrl.SetValue(1)

    def OnMin(self, event):
        min = self.plotWindow.GetAxisRangeMin(self.axisId)
        try:
            newMin = float(self.MinCtrl.GetValue())
        except ValueError:
            wx.Bell()
            self.MinCtrl.SetValue(range.min())
            return

        # Check if the new value is different
        if newMin != min:
            self.plotWindow.SetAxisRange(self.axisId, newMin, self.plotWindow.GetAxisRangeMax(self.axisId))

    def OnMax(self, event):
        max = self.plotWindow.GetAxisRangeMax(self.axisId)
        try:
            newMax = float(self.MaxCtrl.GetValue())
        except ValueError:
            wx.Bell()
            self.MaxCtrl.SetValue(max)
            return

        # Check if the new value is different
        if newMax != max:
            self.plotWindow.SetAxisRange(self.axisId, self.plotWindow.GetAxisRangeMin(self.axisId), newMax)



# This is an example VISAN script for the S5P NO2 L2 product

# Make sure to set the 'products-file directory' option in the VISAN Preferences panel to
# a directory containing S5P NO2 L2 products.

# This example will then take the first product it finds in this directory and
# plot the tropospheric NO2 column


def run():

    import glob
    import wx

    productdir = str(wx.Config.Get().Read('DirectoryLocation/Products'))

    # Use glob to find all S5P NO2 files in productdir
    files = glob.glob(os.path.join(productdir, "S5P_????_L2__NO2*"))

    if len(files) == 0:
        print(("Could not find any S5P NO2 L2 files in directory '" + productdir + "'"))
        return

    # Filter for quality and convert NO2 column number denstiy to Pmolec/cm2 as part of ingestion
    data = harp.import_product(files[0], 'tropospheric_NO2_column_number_density_validity>75;derive(tropospheric_NO2_column_number_density [Pmolec/cm2])')

    wplot(data, value="tropospheric_NO2_column_number_density", colorrange=(0,20), showpropertypanel=True)


run()



# This is an example VISAN script for the GOME-2 L1B product

# Make sure to set the 'products-file directory' option in the VISAN Preferences panel to
# a directory containing GOME-2 L1B products.

# This example will then take the first product it finds in this directory and
# for that product plot the first 100 measured spectra (for the full wavelength range)


def run():

    import glob
    import wx

    productdir = str(wx.Config.Get().Read('DirectoryLocation/Products'))

    # Use glob to find all files in productdir starting with 'GOME_xxx_1B'
    files = glob.glob(os.path.join(productdir, "GOME_xxx_1B*"))
    if len(files) == 0:
        print(("Could not find any GOME_xxx_1B files in directory '" + productdir + "'"))
        return

    # We only ingest the first file from the list
    file = files[0]

    # Get just the geolocation information and cloud fractions (without spectra)
    # We also filter out any backward scan pixels
    geo = harp.import_product(file, 'scan_direction_type=="forward";'
                              'keep(latitude_bounds,longitude_bounds,cloud_fraction)')

    # Since the only 'value' in the geo record is the integration_time the
    # worldmap plot will use the integration_time as color values for each swath
    wplot(geo, title="Cloud fraction from GOME-2 L1B measurements", colorrange=(0, 1))

    # Get the first set of spectra (but only consider the first 200 measurements)
    product = harp.import_product(file, 'scan_direction_type=="forward";index<200')

    plot(product, name=os.path.basename(file), title="GOME-2 L1B spectral readout example")


run()
